! This function conducts a GHK type Monte Carlo integration for each family over SPNEs that rationalize a given location configuration.
! There are four main functions: GHK1, GHK2, GHK3, GHK4.
	! GHK1 is for single-child families, and does not do any monte carlo simulation.
	! GHK2, GHK3, GHK4 are for two-child, three-child, and four-child famlies.
! Three options to choose what to be retruned by this function
! Option 0: return a GHK-simulated probability for the observed location of family m (used for ML estimation)
! Option 1: return a GHK-simulated probability for every location patterns of family m (used to evaluate model fit)
! Option 2: return a GHK-simulated probability for the observed location of family m and simulated error terms for each simulation draw (for counterfactual simulations)
MODULE GHK
	USE Globvar
	IMPLICIT NONE

CONTAINS

FUNCTION GHK1(m,param,n) RESULT(ghkval)
	! This function calculates the GHK simulator for each market*year
USE Globvar
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
INTEGER :: i, j, k, l
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: utilsub, alpha

! PRINT *, 'GHK1-in'

	utilsub=DOT_PRODUCT(param(1:nthetab),xb(:,famindex(m)))
    IF(nthetaa>=3) THEN
    alpha  =DOT_PRODUCT(param(nthetab+1:nthetab+nthetaa-2),xa(:,famindex(m)))
    ELSEIF(nthetaa==1) THEN
    alpha  =param(nthetab+nthetaa)
    ELSE
    alpha = 0.0d0
    ENDIF

	IF (IBITS(near(m),0,1)==1) THEN
    ghkval=1.0d0-NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
	ELSE
    ghkval=NORMCDF(-utilsub-alpha,0.0d0,1.0d0)
    ENDIF


IF (ghkval<=0.0) THEN
!	PRINT *,'Error found!'
!	PRINT *,'GHK Value: ', ghkval
!	PRINT *,'nmkt: ', nmkt
	ghkval=TINY(ghkval)
!	STOP
ENDIF

! PRINT *, 'GHK1-out'

END FUNCTION GHK1



FUNCTION GHK2(m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0,option) RESULT(wgv)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,0:sm2-1,2)
REAL(8), INTENT(IN) :: lchol(2,2)
REAL(8), INTENT(IN) :: xb(nb,2),xa(sa,2),xg(sg,2)
INTEGER, INTENT(IN) :: near0
INTEGER, INTENT(IN) :: option			! compute v (error term) for eldest child if option = 1
LOGICAL :: mask(0:3,2)		! mask(strategy , numchild)	... first child = 1
REAL(8), ALLOCATABLE :: wgv(:,:,:)			! option==0 ... regurn simulated probability based on enter(m)
REAL(8) :: wg(ns,0:sm2-1)
REAL(8) :: v(ns,0:sm2-1,2)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l, nup
INTEGER :: s1, s2
REAL(8) :: temp
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: utilsub(2), alpha(2,3), gamma(2)		! 1 ... oldest child
REAL(8) :: utiltemp

SELECT CASE(option)
CASE(0)
	ALLOCATE (wgv(1,1,1))
	mask = 0
	IF (IBITS(near0,1,1)==1) THEN
		mask(1,1)=1
		nup = 1
	ELSE
		mask(0,1)=1
		nup = 0
	ENDIF
	DO s2=0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,0,1)) mask(s2,2) = 1
	ENDDO
CASE(1)
	ALLOCATE (wgv(0:3,1,1))
	mask = 1
CASE(2)
	ALLOCATE (wgv(ns,0:3,0:2))
	mask = 0
	IF (IBITS(near0,1,1)==1) THEN
		mask(1,1)=1
		nup = 1
	ELSE
		mask(0,1)=1
		nup = 0
	ENDIF
	DO s2=0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,0,1)) mask(s2,2) = 1
	ENDDO
END SELECT


	utilsub(1:2)=MATMUL(param(1:nb),xb(:,1:2))
	IF (na>=3) THEN
	alpha(1:2,1) = MATMUL(param(nb+1:nb+na-2),xa(:,1:2))
	alpha(1:2,2) = MATMUL(param(nb+1:nb+na-2),xa(:,1:2)) * (1.0d0 + param(nb+na-1))
	alpha(1:2,3) = MATMUL(param(nb+1:nb+na-2),xa(:,1:2)) * (1.0d0 + param(nb+na-1) + param(nb+na))
	ELSEIF (na==1) THEN
	alpha(1:2,1:3)=param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:2) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:2))
	ELSEIF (ng==1) THEN
	gamma(1:2)=param(nb+na+ng)
	ELSE
	gamma=0.0d0
	ENDIF


wgv=0.000d0
wg=0.000d0
v=0.000d0

! One way to construct a simulated probability for a location configuration is to
! simulate a probability for each SPNE that rationalizes the observed location configuration.
! But the algorithm below takes a different approach: Evaluate the probabilities of all 
! strategies for the youngest child first, and then the second-youngest child, to the eldest child.
! This makes the computation algorithm simpler and faster.

IF (mask(3,2)==1) THEN	! s2==3
! condition: epsiron > max(-du0, -du1)
	ta=NORMCDF((max(-utilsub(2)-alpha(2,1),-utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2))),0.0d0,1.0d0)

	wg(:,3)=1.0-ta

	DO j=1, ns
		IF (ta(j)>=0.99999999d0) THEN 
!	        PRINT *,'Warning: a is too large.'
			v(j,3,2)=HUGE(v(j,3,2))
		ELSE
        	v(j,3,2)=INVNORMCDF((ta(j)*(1-draws(j,3,2))+draws(j,3,2)),0.0d0,1.0d0)
		ENDIF
	END DO

ENDIF

IF (mask(0,2)==1) THEN	! s2==0
! condition: epsiron < min(-du0, -du1)
	tb=NORMCDF(min(-utilsub(2)-alpha(2,1),-utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2)),0.0d0,1.0d0)
	wg(:,0)=tb

	DO j=1, ns
	   	IF (tb(j)<=0.00000001d0) THEN 
	!		PRINT *,'Warning: b is too small.'
			v(j,0,2)=-HUGE(v(j,0,2))
		ELSE
		v(j,0,2)=INVNORMCDF((tb(j)*draws(j,0,2)),0.0d0,1.0d0)
		ENDIF
	END DO

ENDIF

IF (mask(1,2)==1) THEN       ! s2==1
! condition: -du0 < epsiron < -du1
	IF ((-utilsub(2)-alpha(2,1)) < -utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2)) THEN
		ta=NORMCDF((-utilsub(2)-alpha(2,1)),0.0d0,1.0d0)
		tb=NORMCDF((-utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2)),0.0d0,1.0d0)
		wg(:,1)=tb-ta

		DO j=1, ns
		   	IF (tb(j)<=0.00000001d0) THEN 
		!		    PRINT *,'Warning: b is too small.'
				v(j,1,2)=-HUGE(v(j,1,2))
			ELSE
		  	v(j,1,2)=INVNORMCDF((ta(j)*(1-draws(j,1,2))+tb(j)*draws(j,1,2)),0.0d0,1.0d0)
			ENDIF
		END DO

	ENDIF

ENDIF

IF (mask(2,2)==1) THEN		! s2==2
! condition: -du1 < epsiron < -du0
	IF ((-utilsub(2)-alpha(2,1)) > -utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2)) THEN
		ta=NORMCDF((-utilsub(2)-(alpha(2,2)-alpha(2,1))-gamma(2)),0.0d0,1.0d0)
		tb=NORMCDF((-utilsub(2)-alpha(2,1)),0.0d0,1.0d0)
		wg(:,2)=tb-ta

		DO j=1, ns
		   	IF (tb(j)<=0.00000001d0) THEN 
		!		    PRINT *,'Warning: b is too small.'
				v(j,2,2)=-HUGE(v(j,2,2))
			ELSE
		  	v(j,2,2)=INVNORMCDF((ta(j)*(1-draws(j,2,2))+tb(j)*draws(j,2,2)),0.0d0,1.0d0)
			ENDIF
		END DO

	ENDIF
ENDIF



! player 1

DO s2=0,3			! Strategy loop s2=0,2**2-1
!IF (mask(s2,2)==1 .and. sum(wg(:,s2))>0.0d0) THEN	! This change does not make any difference
IF (mask(s2,2)==1 .and. wg(1,s2)>0.0d0) THEN

	utiltemp=utilsub(1)+ &
		IBITS(s2,1,1)*alpha(1,2)+ &
		(1-IBITS(s2,1,1))*alpha(1,1)+ &
		IBITS(s2,1,1)*gamma(1)
	utiltemp=utiltemp-(IBITS(s2,0,1))*alpha(1,1)

	temp=1.0d0/lchol(2,2)

	IF (mask(1,1)==1) THEN
		k=2+IBITS(s2,1,1)
		DO j=1, ns
			ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s2,2:2)))*temp,0.0d0,1.0d0)
			IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,s2)*(1.0-ta(j))
			IF (option==2) wgv(j,s2,1) = INVNORMCDF((ta(j)*(1-draws(j,s2,1))+draws(j,s2,1)),0.0d0,1.0d0)
		END DO         ! Simulation loop
		IF (option==0) wg(:,s2)=wg(:,s2)*(1.0-ta)
		IF (option==2) wgv(:,s2,0)=wg(:,s2)*(1.0-ta)
	ENDIF
	IF (mask(0,1)==1) THEN
		k=IBITS(s2,0,1)
		DO j=1, ns
			tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(2,1:1),v(j,s2,2:2)))*temp,0.0d0,1.0d0)
			IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,s2)*tb(j)
			IF (option==2) wgv(j,s2,1) = INVNORMCDF((tb(j)*draws(j,s2,1)),0.0d0,1.0d0)
		END DO         ! Simulation loop
		IF (option==0) wg(:,s2)=wg(:,s2)*tb
		IF (option==2) wgv(:,s2,0)=wg(:,s2)*tb
    ENDIF
ENDIF
END DO         ! Strategy loop

IF (option==0) THEN
	wgv(1,1,1)=SUM(wg(:,:))	! wg(ns,s2*128+s3*16+s4)
	wgv=wgv/ns

	IF (wgv(1,1,1)<=0.0) THEN
	!	PRINT *,'Error found!'
	!	PRINT *,'GHK Value: ', ghkval
		wgv(1,1,1)=TINY(wgv(1,1,1))
	!	STOP
	ENDIF
END IF

IF (option==1) THEN
	wgv=wgv/ns
ENDIF

IF (option==2) THEN
	wgv(:,:,2)=v(:,:,2)
ENDIF

END FUNCTION GHK2




FUNCTION GHK3(m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0,option) RESULT(wgv)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,0:sm3-1,3)
REAL(8), INTENT(IN) :: lchol(3,3)
REAL(8), INTENT(IN) :: xb(nb,3),xa(sa,3),xg(sg,3)
INTEGER, INTENT(IN) :: near0
INTEGER, INTENT(IN) :: option			! compute v (error term) for eldest child if option = 1
LOGICAL :: mask(0:7,3)		! mask(strategy , numchild)	... first child = 1
REAL(8), ALLOCATABLE :: wgv(:,:,:)			! option==0 ... regurn simulated probability based on enter(m)
REAL(8) :: wg(ns,0:sm3-1)
REAL(8) :: v(ns,0:sm3-1,3)   ! GHK simulation draws constructed from draws()

INTEGER :: i, j, k, l, nup
INTEGER :: s1, s2, s3, stmp
REAL(8) :: temp
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: utilsub(3), alpha(3,3), gamma(3,2)		! 1 ... oldest child
REAL(8) :: utiltemp

SELECT CASE(option)
CASE(0)
	ALLOCATE (wgv(1,1,1))
	mask = 0
	IF (IBITS(near0,2,1)==1) THEN
		mask(1,1) = 1
		nup = 1
	ELSE
		mask(0,1) = 1
		nup = 0
	ENDIF
	DO s2 = 0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,1,1)) mask(s2,2) = 1
	ENDDO
	nup = nup+IBITS(near0,1,1)
	DO s3 = 0,7
		IF (IBITS(s3,nup,1)==IBITS(near0,0,1)) mask(s3,3) = 1
	ENDDO
CASE(1)
	ALLOCATE (wgv(0:7,1,1))
	mask = 1
CASE(2)
	ALLOCATE (wgv(ns,0:31,0:3))
	mask = 0
	IF (IBITS(near0,2,1)==1) THEN
		mask(1,1) = 1
		nup = 1
	ELSE
		mask(0,1) = 1
		nup = 0
	ENDIF
	DO s2 = 0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,1,1)) mask(s2,2) = 1
	ENDDO
	nup = nup+IBITS(near0,1,1)
	DO s3 = 0,7
		IF (IBITS(s3,nup,1)==IBITS(near0,0,1)) mask(s3,3) = 1
	ENDDO
END SELECT




	utilsub(1:3)=MATMUL(param(1:nb),xb(:,1:3))
	IF (na>=3) THEN
	alpha(1:3,1) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) )
	alpha(1:3,2) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) ) * ( 1.0d0 + param(nb+na-1) )
	alpha(1:3,3) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) ) * ( 1.0d0 + param(nb+na-1) + param(nb+na) )
	ELSEIF (na==1) THEN
	alpha(1:3,1:3)=param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:3,1) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:3))
	gamma(1:3,2) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:3))*(1.0d0+param(nb+na+ng))
	ELSEIF (ng==1) THEN
	gamma(1:3,1:2)=param(nb+na+ng)
	ELSE
	gamma=0.0d0
	ENDIF


wg=0.000d0
v=0.000d0
wgv=0.000d0

IF (mask(0,3)==1) THEN		! s3==0
	s3=0

	! condition: epsiron < min(-du0, -du1, -du2)
	tb=NORMCDF(min(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
	DO s2=0,3	! s2=0,2**2-1
		stmp=s2*8+s3
		IF (mask(s2,2)==1) THEN
			wg(:,stmp)=tb
			DO j=1, ns
				IF (tb(j)<=0.00000001d0) THEN 
				!    PRINT *,'Warning: b is too small.'
        			v(j,stmp,3)=-HUGE(v(j,stmp,3))
   				ELSE
   					v(j,stmp,3)=INVNORMCDF((tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
				ENDIF
			END DO
		ENDIF
	END DO
ENDIF

IF (mask(1,3)==1) THEN		! s3==1
	s3=1

	! condition: -du0 < epsiron < min(-du1, -du2)
	IF (-utilsub(3)-alpha(3,1) < min(-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2))) THEN
		ta=NORMCDF((-utilsub(3)-alpha(3,1)),0.0d0,1.0d0)
		tb=NORMCDF(min(-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
				wg(:,stmp)=tb-ta

				DO j=1, ns
					IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,3)=-HUGE(v(j,stmp,3))
   					ELSE
					v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
					ENDIF
				END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF

IF (mask(2,3)==1) THEN		! s3==2
	s3=2
	! condition: -du1 < epsiron < min(-du0, -du2)
	IF (-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1) < min(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2))) THEN
		ta=NORMCDF((-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1)),0.0d0,1.0d0)
		tb=NORMCDF(min(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
				wg(:,stmp)=tb-ta

				DO j=1, ns
					IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,3)=-HUGE(v(j,stmp,3))
   					ELSE
					v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
					ENDIF
				END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF

IF (mask(3,3)==1) THEN		! s3==3
	s3=3
	! condition: max(-du0, -du1) < epsiron < -du2
	IF (max(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1)) < -utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)) THEN
		ta=NORMCDF(max(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1)),0.0d0,1.0d0)
		tb=NORMCDF(-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
				wg(:,stmp)=tb-ta

				DO j=1, ns
					IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,3)=-HUGE(v(j,stmp,3))
   					ELSE
					v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
					ENDIF
				END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF

IF (mask(4,3)==1) THEN		! s3==4
	s3=4
	! condition: -du2 < epsiron < min(-du0, -du1)
	IF (-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2) < min(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1))) THEN
		ta=NORMCDF((-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
		tb=NORMCDF(min(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1)),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
			wg(:,stmp)=tb-ta

			DO j=1, ns
				IF (tb(j)<=0.00000001d0) THEN 
		!		    PRINT *,'Warning: b is too small.'
        			v(j,stmp,3)=-HUGE(v(j,stmp,3))
   				ELSE
				v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
				ENDIF
			END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF
    
IF (mask(5,3)==1) THEN		! s3==5
	s3=5
	! condition: max(-du0, -du2) < epsiron < -du1
	IF (max(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)) < -utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1)) THEN
		ta=NORMCDF(max(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
		tb=NORMCDF(-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
			wg(:,stmp)=tb-ta

			DO j=1, ns
				IF (tb(j)<=0.00000001d0) THEN 
		!		    PRINT *,'Warning: b is too small.'
        			v(j,stmp,3)=-HUGE(v(j,stmp,3))
   				ELSE
				v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
				ENDIF
			END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF


IF (mask(6,3)==1) THEN		! s3==6
	s3=6
	! condition: max(-du1, -du2) < epsiron < -du0
	IF (max(-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)) < -utilsub(3)-alpha(3,1)) THEN
		ta=NORMCDF(max(-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2)),0.0d0,1.0d0)
		tb=NORMCDF((-utilsub(3)-alpha(3,1)),0.0d0,1.0d0)
		DO s2=0,3	! s2=0,2**2-1
			stmp=s2*8+s3
			IF (mask(s2,2)==1) THEN
				wg(:,stmp)=tb-ta

				DO j=1, ns
					IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,3)=-HUGE(v(j,stmp,3))
   					ELSE
					v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
					ENDIF
				END DO
			ENDIF
		ENDDO
	ENDIF
ENDIF

IF (mask(7,3)==1) THEN
	s3=7
	! condition: epsiron > max(-du0, -du1, -du2)
	ta=NORMCDF((max(-utilsub(3)-alpha(3,1),-utilsub(3)-(alpha(3,2)-alpha(3,1))-gamma(3,1),-utilsub(3)-(alpha(3,3)-alpha(3,2))-gamma(3,2))),0.0d0,1.0d0)

	DO s2=0,3	! s2=0,2**2-1
		stmp=s2*8+s3
		IF (mask(s2,2)==1) THEN
			wg(:,stmp)=1.0-ta
			DO j=1, ns
       			IF (ta(j)>=0.99999999d0) THEN 
		!	            PRINT *,'Warning: a is too large.'
    				v(j,stmp,3)=HUGE(v(j,stmp,3))
				ELSE
					v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+draws(j,stmp,3)),0.0d0,1.0d0)
				ENDIF
			END DO
		ENDIF
	END DO
ENDIF




temp=1.0d0/lchol(2,2)
DO s2=0,2**2-1       ! Strategy loop
IF (mask(s2,2)==1) THEN
	DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
			stmp=s2*8+s3

			IF (s2==0) THEN

    			DO j=1, ns
				tb(j)=NORMCDF((min( &
					-(utilsub(2) + IBITS(s3,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*alpha(2,1) + IBITS(s3,1,1)*gamma(2,1) - IBITS(s3,0,1)*alpha(2,1) ),   &   ! when p1 far
					-(utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*alpha(2,2) + IBITS(s3,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*gamma(2,1) - IBITS(s3,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*alpha(2,1) )  &   ! when p1 near
					) &
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
    			END DO         ! Simulation loop
				wg(:,stmp)=wg(:,stmp)*tb

				DO j=1, ns
           			IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,2)=-HUGE(v(j,stmp,2))
   					ELSE
   					v(j,stmp,2)=INVNORMCDF((tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
					ENDIF
				END DO
    
			ELSE IF (s2==3) THEN

    			DO j=1, ns
				ta(j)=NORMCDF((max( &
					-(utilsub(2) + IBITS(s3,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*alpha(2,1) + IBITS(s3,1,1)*gamma(2,1) - IBITS(s3,0,1)*alpha(2,1)),   &   ! when p1 far
					-(utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*alpha(2,2) + IBITS(s3,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*gamma(2,1) - IBITS(s3,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*alpha(2,1) )  &   ! when p1 near
					) &
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
    			END DO         ! Simulation loop
				wg(:,stmp)=wg(:,stmp)*(1.0d0-ta)

				DO j=1, ns
       				IF (ta(j)>=0.99999999d0) THEN 
		!	            PRINT *,'Warning: a is too large.'
    					v(j,stmp,2)=HUGE(v(j,stmp,2))
					ELSE
            			v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+draws(j,stmp,2)),0.0d0,1.0d0)
					ENDIF
				END DO


			ELSE IF (s2==1) THEN       ! s2==1
    			DO j=1, ns
				ta(j)=NORMCDF(( &
					-(utilsub(2) + IBITS(s3,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*alpha(2,1) + IBITS(s3,1,1)*gamma(2,1) - IBITS(s3,0,1)*alpha(2,1))   &   ! when p1 far
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
				tb(j)=NORMCDF(( &
					-(utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*alpha(2,2) + IBITS(s3,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*gamma(2,1) - IBITS(s3,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*alpha(2,1) )  &   ! when p1 near
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
    			END DO         ! Simulation loop

				DO j=1, ns
					IF (tb(j)-ta(j) > 0.0d0) THEN
					wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       				IF (ta(j)>=0.99999999d0) THEN 
		!	            PRINT *,'Warning: a is too large.'
    					v(j,stmp,2)=HUGE(v(j,stmp,2))
					ELSE IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,2)=-HUGE(v(j,stmp,2))
   					ELSE
          				v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
					ENDIF
					ELSE
						wg(j,stmp)=0.0d0
					ENDIF
				END DO

			ELSE IF (s2==2) THEN       ! s2==2
    			DO j=1, ns
				ta(j)=NORMCDF(( &
					-(utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*alpha(2,2) + IBITS(s3,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*gamma(2,1) - IBITS(s3,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*alpha(2,1) )  &   ! when p1 near
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
				tb(j)=NORMCDF(( &
					-(utilsub(2) + IBITS(s3,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*alpha(2,1) + IBITS(s3,1,1)*gamma(2,1) - IBITS(s3,0,1)*alpha(2,1) )   &   ! when p1 far
				-DOT_PRODUCT(lchol(2,1:1),v(j,stmp,3:3)))*temp,0.0d0,1.0d0)
    			END DO         ! Simulation loop

				DO j=1, ns
					IF (tb(j)-ta(j) > 0.0d0) THEN
					wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       				IF (ta(j)>=0.99999999d0) THEN 
		!	            PRINT *,'Warning: a is too large.'
    					v(j,stmp,2)=HUGE(v(j,stmp,2))
					ELSE IF (tb(j)<=0.00000001d0) THEN 
			!		    PRINT *,'Warning: b is too small.'
        				v(j,stmp,2)=-HUGE(v(j,stmp,2))
   					ELSE
          				v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
					ENDIF
					ELSE
						wg(j,stmp)=0
					ENDIF
				END DO

			ENDIF


		ELSE
			wg(:,s2*8+s3)=0.0d0
		ENDIF
	END DO         ! Strategy loop
ELSE
    DO s3=0,2**3-1
    wg(:,s2*8+s3)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop






! player 1

DO s2=0,3       ! Strategy loop	s2=0,2**2-1
IF (mask(s2,2)==1) THEN
	DO s3=0,7       ! Strategy loop	s3=0,2**3-1
		stmp=s2*8+s3
		IF (mask(s3,3)==1 .and. wg(1,stmp)>0.0d0) THEN

			utiltemp=utilsub(1) &
				+ IBITS(s2,1,1)*IBITS(s3,2,1)*alpha(1,3) + IBITS(s2,1,1)*(1-IBITS(s3,2,1))*alpha(1,2) + (1-IBITS(s2,1,1))*IBITS(s3,1,1)*alpha(1,2) + (1-IBITS(s2,1,1))*(1-IBITS(s3,1,1))*alpha(1,1) &
				+ IBITS(s2,1,1)*IBITS(s3,2,1)*gamma(1,2) + IBITS(s2,1,1)*(1-IBITS(s3,2,1))*gamma(1,1) + (1-IBITS(s2,1,1))*IBITS(s3,1,1)*gamma(1,1)
			IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==1) THEN
					utiltemp=utiltemp-alpha(1,2)
				ELSE IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==0) THEN
					utiltemp=utiltemp-alpha(1,1)
				ELSE IF (IBITS(s2,0,1)==0 .and. IBITS(s3,0,1)==1) THEN
					utiltemp=utiltemp-alpha(1,1)
			ENDIF

			temp=1.0d0/lchol(3,3)
			IF (mask(1,1)==1) THEN
				k=4+IBITS(s2,1,1)*2+IBITS(s3,1+IBITS(s2,1,1),1)
				DO j=1, ns
					ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,stmp,3:2:-1)))*temp,0.0d0,1.0d0)
					IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,stmp)*(1.0-ta(j))
					IF (option==2) wgv(j,stmp,1) = INVNORMCDF((ta(j)*(1-draws(j,stmp,1))+draws(j,stmp,1)),0.0d0,1.0d0)
				END DO         ! Simulation loop
				IF (option==0) wg(:,stmp)=wg(:,stmp)*(1.0-ta)
				IF (option==2) wgv(:,stmp,0)=wg(:,stmp)*(1.0-ta)
			ENDIF
			IF (mask(0,1)==1) THEN
				k=IBITS(s2,0,1)*2+IBITS(s3,IBITS(s2,0,1),1)
				DO j=1, ns
					tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(3,1:2),v(j,stmp,3:2:-1)))*temp,0.0d0,1.0d0)
					IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,stmp) * tb(j)
					IF (option==2) wgv(j,stmp,1) = INVNORMCDF((tb(j)*draws(j,stmp,1)),0.0d0,1.0d0)
				END DO         ! Simulation loop
				IF (option==0) wg(:,stmp)=wg(:,stmp)*tb
				IF (option==2) wgv(:,stmp,0)=wg(:,stmp)*tb
			ENDIF

		ENDIF
	END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop


IF (option==0) THEN
	wgv(1,1,1)=SUM(wg(:,:))	! wg(ns,s2*128+s3*16+s4)
	wgv=wgv/ns

	IF (wgv(1,1,1)<=0.0) THEN
	!	PRINT *,'Error found!'
	!	PRINT *,'GHK Value: ', ghkval
		wgv(1,1,1)=TINY(wgv(1,1,1))
	!	STOP
	ENDIF
END IF

IF (option==1) THEN
	wgv=wgv/ns
ENDIF

IF (option==2) THEN
	wgv(:,:,2)=v(:,:,2)
	wgv(:,:,3)=v(:,:,3)
ENDIF


END FUNCTION GHK3




FUNCTION GHK4(m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0,option) RESULT(wgv)
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,0:sm4-1,4)
REAL(8), INTENT(IN) :: lchol(4,4)
REAL(8), INTENT(IN) :: xb(nb,4),xa(sa,4),xg(sg,4)
INTEGER, INTENT(IN) :: near0
INTEGER, INTENT(IN) :: option			! compute v (error term) for eldest child if option = 1
LOGICAL :: mask(0:15,4)		! mask(strategy , numchild)	... first child = 1
REAL(8), ALLOCATABLE :: wgv(:,:,:)			! option==0 ... regurn simulated probability based on enter(m)
REAL(8) :: wg(ns,0:sm4-1)
REAL(8) :: v(ns,0:sm4-1,4)   ! GHK simulation draws constructed from draws()
INTEGER :: i, j, k, l, nup
INTEGER :: s1, s2, s3, s4, stmp
REAL(8) :: temp
REAL(8) :: ta(ns), tb(ns)
REAL(8) :: utilsub(4), alpha(4,3), gamma(4,3)		! 1 ... oldest child
REAL(8) :: utiltemp

SELECT CASE(option)
CASE(0)
	ALLOCATE (wgv(1,1,1))
	mask=0
	IF (IBITS(near0,3,1)==1) THEN
		mask(1,1)=1
		nup=1
	ELSE
		mask(0,1)=1
		nup=0
	ENDIF
	DO s2 = 0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,2,1)) mask(s2,2)=1
	ENDDO
	nup = nup+IBITS(near0,2,1)
	DO s3 = 0,7
		IF (IBITS(s3,nup,1)==IBITS(near0,1,1)) mask(s3,3)=1
	ENDDO
	nup = nup+IBITS(near0,1,1)
	DO s4 = 0,15
		IF (IBITS(s4,nup,1)==IBITS(near0,0,1)) mask(s4,4)=1
	ENDDO
CASE(1)
	ALLOCATE (wgv(0:15,1,1))
	mask=1
CASE(2)
	ALLOCATE (wgv(ns,0:511,0:4))
	mask=0
	IF (IBITS(near0,3,1)==1) THEN
		mask(1,1)=1
		nup=1
	ELSE
		mask(0,1)=1
		nup=0
	ENDIF
	DO s2 = 0,3
		IF (IBITS(s2,nup,1)==IBITS(near0,2,1)) mask(s2,2)=1
	ENDDO
	nup = nup+IBITS(near0,2,1)
	DO s3 = 0,7
		IF (IBITS(s3,nup,1)==IBITS(near0,1,1)) mask(s3,3)=1
	ENDDO
	nup = nup+IBITS(near0,1,1)
	DO s4 = 0,15
		IF (IBITS(s4,nup,1)==IBITS(near0,0,1)) mask(s4,4)=1
	ENDDO
END SELECT



	utilsub(1:4)=MATMUL(param(1:nb),xb(:,1:4))
	IF (na>=3) THEN
    alpha(1:4,1) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) )
    alpha(1:4,2) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) ) * ( 1.0d0 + param(nb+na-1) )
    alpha(1:4,3) = ( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) ) * ( 1.0d0 + param(nb+na-1) + param(nb+na) )
	ELSEIF (na==1) THEN
	alpha(1:4,1:3) = param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:4,1) = MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:4))
	gamma(1:4,2) = MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:4))*(1.0d0+param(nb+na+ng))
	gamma(1:4,3) = MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:4))*(1.0d0+param(nb+na+ng)*2.0d0)
	ELSEIF (ng==1) THEN
	gamma(1:4,1:3)=param(nb+na+ng)
	ELSE
	gamma = 0.0d0
	ENDIF


wgv = 0.0d0
wg = 0.0d0
v = 0.0d0


	IF (mask(0,4)==1) THEN
        s4=0
        ! condition: epsiron < min(-du0, -du1, -du2, -du3)
			! du ... util(far)-util(near) ==>	-du = u(near) - u(far)
        tb=NORMCDF(min( & 
			-utilsub(4)-alpha(4,1), &	! -du0
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &	! -du1
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &	! -du2
			-utilsub(4)-gamma(4,3) &		! -du3
			),0.0d0,1.0d0)

        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
   	            v(j,stmp,4)=INVNORMCDF((tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        END DO
        ENDIF
        END DO

    ENDIF

	IF (mask(1,4)==1) THEN		! s4==1
        s4=1

        ! condition: -du0 < epsiron < min(-du1, -du2, -du3)
        IF (-utilsub(4)-alpha(4,1) < min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			)) THEN
        ta=NORMCDF((-utilsub(4)-alpha(4,1)),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)

        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(2,4)==1) THEN		! s4==2
        s4=2

        ! condition: -du1 < epsiron < min(-du0, -du2, -du3)
        IF (-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			)) THEN
        ta=NORMCDF(-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(3,4)==1) THEN		! s4==3
        s4=3

        ! condition: max(-du0, -du1) < epsiron < min(-du2, -du3)
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1) &
		) < min( &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
		)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(4,4)==1) THEN		! s4==4
        s4=4

        ! condition: -du2 < epsiron < min(-du0, -du1, -du3)
        IF (-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			)) THEN
        ta=NORMCDF(-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(5,4)==1) THEN		! s4==5
        s4=5

        ! condition: max(-du0, -du2) < epsiron < min(-du1, -du3)
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
		) < min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
		)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(6,4)==1) THEN		! s4==6
        s4=6

        ! condition: max(-du1, -du2) < epsiron < min(-du0, -du3)
        IF (max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
		) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-gamma(4,3) &
		)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(7,4)==1) THEN		! s4==7
        s4=7

        ! condition: max(-du0, -du1, -du2) < epsiron < -du3
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			) < -utilsub(4)-gamma(4,3)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        tb=NORMCDF(-utilsub(4)-gamma(4,3),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(8,4)==1) THEN		! s4==8
        s4=8

        ! condition: -du3 < epsiron < min(-du0, -du1, -du2)
        IF (-utilsub(4)-gamma(4,3) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			)) THEN
        ta=NORMCDF(-utilsub(4)-gamma(4,3),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(9,4)==1) THEN		! s4==9
        s4=9

        ! condition: max(-du0, -du3) < epsiron < min(-du1, -du2)
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-gamma(4,3) &
			) < min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(10,4)==1) THEN		! s4==10
        s4=10

        ! condition: max(-du1, -du3) < epsiron < min(-du0, -du2)
        IF (max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(11,4)==1) THEN		! s4==11
        s4=11

        ! condition: max(-du0, -du1, -du3) < epsiron < -du2
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			) < -utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF(-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(12,4)==1) THEN		! s4==12
        s4=12

        ! condition: max(-du2, -du3) < epsiron < min(-du0, -du1)
        IF (max( &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			) < min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1) &
			)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF(min( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1) &
			),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(13,4)==1) THEN		! s4==13
        s4=13

        ! condition: max(-du0, -du2, -du3) < epsiron < -du1
        IF (max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			) < -utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF(-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(14,4)==1) THEN		! s4==14
        s4=14

        ! condition: max(-du1, -du2, -du3) < epsiron < -du0
        IF (max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			) < -utilsub(4)-alpha(4,1)) THEN
        ta=NORMCDF(max( &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			),0.0d0,1.0d0)
        tb=NORMCDF((-utilsub(4)-alpha(4,1)),0.0d0,1.0d0)
        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=tb-ta

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,4)=-HUGE(v(j,stmp,4))
   	        ELSE
          	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+tb(j)*draws(j,stmp,4)),0.0d0,1.0d0)
            ENDIF
        END DO
        ENDIF
        ENDDO
        ENDIF
        ENDDO
        ENDIF

    ENDIF

	IF (mask(15,4)==1) THEN
        s4=15
        
        ! condition: epsiron > max(-du0, -du1, -du2, -du3)
        ta=NORMCDF((max( &
			-utilsub(4)-alpha(4,1), &
			-utilsub(4)-(alpha(4,2)-alpha(4,1))-gamma(4,1), &
			-utilsub(4)-(alpha(4,3)-alpha(4,2))-gamma(4,2), &
			-utilsub(4)-gamma(4,3) &
			)),0.0d0,1.0d0)

        DO s2=0,2**2-1
		IF (mask(s2,2)==1) THEN
        DO s3=0,2**3-1
		IF (mask(s3,3)==1) THEN
		stmp=s2*128+s3*16+s4
        wg(:,stmp)=1.0-ta

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,4)=HUGE(v(j,stmp,4))
            ELSE
            	v(j,stmp,4)=INVNORMCDF((ta(j)*(1-draws(j,stmp,4))+draws(j,stmp,4)),0.0d0,1.0d0)
	        ENDIF
        END DO
        ENDIF
        END DO
        ENDIF
        END DO

    ENDIF




temp=1.0d0/lchol(2,2)
DO s2=0,2**2-1       ! Strategy loop
IF (mask(s2,2)==1) THEN
DO s3=0,2**3-1
IF (mask(s3,3)==1) THEN
DO s4=0,2**4-1
IF (mask(s4,4)==1) THEN
	stmp=s2*128+s3*16+s4

    IF (s3==0) THEN
    	DO j=1, ns
        tb(j)=NORMCDF((min( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1)),  &   ! when upstream 1
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,stmp)=wg(:,stmp)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
!            PRINT *,'k,s1,s2',k,s1,s2
   	        v(j,stmp,3)=INVNORMCDF((tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
        END DO

    ELSE IF (s3==1) THEN       ! s3==1
    	DO j=1, ns
        ta(j)=NORMCDF(( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1))   &   ! when upstraem 0
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF((min( &
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1)),  &   ! when upstream 1
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==2) THEN       ! s3==2
    	DO j=1, ns
        ta(j)=NORMCDF(( &
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1))  &   ! when upstream 1
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF((min( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==3) THEN       ! s3==3
    	DO j=1, ns
        ta(j)=NORMCDF((max( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1))  &   ! when upstream 1
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF(( &
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==4) THEN       ! s3==4
    	DO j=1, ns
        ta(j)=NORMCDF(( &
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF((min( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1))  &   ! when upstream 1
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==5) THEN       ! s3==5
    	DO j=1, ns
        ta(j)=NORMCDF((max( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF(( &
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1))  &   ! when upstream 1
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==6) THEN       ! s3==6
    	DO j=1, ns
        ta(j)=NORMCDF((max( &
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1)),  &   ! when upstream 1
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF(( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1))   &   ! when upstraem 0
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,3)=-HUGE(v(j,stmp,3))
   	        ELSE
          	    v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+tb(j)*draws(j,stmp,3)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s3==7) THEN
    	DO j=1, ns
        ta(j)=NORMCDF((max( &
			-(utilsub(3) + IBITS(s4,1,1)*alpha(3,2) + (1-IBITS(s4,1,1))*alpha(3,1) + IBITS(s4,1,1)*gamma(3,1) - IBITS(s4,0,1)*alpha(3,1)),   &   ! when upstraem 0
			-(utilsub(3) + IBITS(s4,2,1)*alpha(3,3) + (1-IBITS(s4,2,1))*alpha(3,2) + IBITS(s4,2,1)*gamma(3,2) + (1-IBITS(s4,2,1))*gamma(3,1) - IBITS(s4,1,1)*alpha(3,2) - (1-IBITS(s4,1,1))*alpha(3,1)),  &   ! when upstream 1
			-( utilsub(3) + alpha(3,3) + IBITS(s4,3,1)*gamma(3,3) + (1-IBITS(s4,3,1))*gamma(3,2) - IBITS(s4,2,1)*alpha(3,3) - (1-IBITS(s4,2,1))*alpha(3,2) )  &   ! when upstream 2
            ) &
        -DOT_PRODUCT(lchol(2,1:1),v(j,stmp,4:4)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,stmp)=wg(:,stmp)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,3)=HUGE(v(j,stmp,3))
            ELSE
            	v(j,stmp,3)=INVNORMCDF((ta(j)*(1-draws(j,stmp,3))+draws(j,stmp,3)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE

    	wg(:,stmp)=0.0d0
    	v(:,stmp,3)=0.0d0
        
    ENDIF


ELSE
    wg(:,s2*128+s3*16+s4)=0.0d0
ENDIF
END DO         ! Strategy loop
ELSE
    DO s4=0,2**4-1
    wg(:,s2*128+s3*16+s4)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop
ELSE
    DO s3=0,2**3-1
    DO s4=0,2**4-1
    wg(:,s2*128+s3*16+s4)=0.0d0
    END DO
    END DO
ENDIF
END DO         ! Strategy loop



temp=1.0d0/lchol(3,3)
DO s2=0,2**2-1       ! Strategy loop
IF (mask(s2,2)==1) THEN
DO s3=0,2**3-1
IF (mask(s3,3)==1) THEN
DO s4=0,2**4-1
IF (mask(s4,4)==1) THEN
	stmp=s2*128+s3*16+s4

    IF (s2==0) THEN
    	DO j=1, ns
        tb(j)=NORMCDF((min( &
			-( utilsub(2) + IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) + IBITS(s3,1,1)*IBITS(s4,2,1)*gamma(2,2) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*gamma(2,1) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*gamma(2,1) - IBITS(s3,0,1)*IBITS(s4,1,1)*alpha(2,2) - IBITS(s3,0,1)*(1-IBITS(s4,1,1))*alpha(2,1) - (1-IBITS(s3,0,1))*IBITS(s4,0,1)*alpha(2,1) ),   &   ! when upstraem 0
			-( utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*alpha(2,2) + IBITS(s3,2,1)*IBITS(s4,3,1)*gamma(2,3) + IBITS(s3,2,1)*(1-IBITS(s4,3,1))*gamma(2,2) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*gamma(2,1) - IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) - IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) - (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) )  &   ! when upstream 1
            ) &
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,stmp)=wg(:,stmp)*tb

        DO j=1, ns
           	IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,2)=-HUGE(v(j,stmp,2))
   	        ELSE
   	        v(j,stmp,2)=INVNORMCDF((tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
            ENDIF
        END DO
    
    ELSE IF (s2==3) THEN
    	DO j=1, ns
        ta(j)=NORMCDF((max( &
			-( utilsub(2) + IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) + IBITS(s3,1,1)*IBITS(s4,2,1)*gamma(2,2) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*gamma(2,1) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*gamma(2,1) - IBITS(s3,0,1)*IBITS(s4,1,1)*alpha(2,2) - IBITS(s3,0,1)*(1-IBITS(s4,1,1))*alpha(2,1) - (1-IBITS(s3,0,1))*IBITS(s4,0,1)*alpha(2,1) ),   &   ! when upstraem 0
			-( utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*alpha(2,2) + IBITS(s3,2,1)*IBITS(s4,3,1)*gamma(2,3) + IBITS(s3,2,1)*(1-IBITS(s4,3,1))*gamma(2,2) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*gamma(2,1) - IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) - IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) - (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) )  &   ! when upstream 1
            ) &
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop
        wg(:,stmp)=wg(:,stmp)*(1.0d0-ta)

        DO j=1, ns
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,2)=HUGE(v(j,stmp,2))
            ELSE
            	v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+draws(j,stmp,2)),0.0d0,1.0d0)
	        ENDIF
        END DO


    ELSE IF (s2==1) THEN       ! s2==1
    	DO j=1, ns
        ta(j)=NORMCDF(( &
			-( utilsub(2) + IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) + IBITS(s3,1,1)*IBITS(s4,2,1)*gamma(2,2) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*gamma(2,1) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*gamma(2,1) - IBITS(s3,0,1)*IBITS(s4,1,1)*alpha(2,2) - IBITS(s3,0,1)*(1-IBITS(s4,1,1))*alpha(2,1) - (1-IBITS(s3,0,1))*IBITS(s4,0,1)*alpha(2,1) )   &   ! when upstraem 0
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF(( &
			-( utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*alpha(2,2) + IBITS(s3,2,1)*IBITS(s4,3,1)*gamma(2,3) + IBITS(s3,2,1)*(1-IBITS(s4,3,1))*gamma(2,2) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*gamma(2,1) - IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) - IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) - (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) )  &   ! when upstream 1
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop


        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,2)=HUGE(v(j,stmp,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,2)=-HUGE(v(j,stmp,2))
   	        ELSE
          	    v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE IF (s2==2) THEN       ! s2==2
    	DO j=1, ns
        ta(j)=NORMCDF(( &
			-( utilsub(2) + IBITS(s3,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*alpha(2,3) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*alpha(2,2) + IBITS(s3,2,1)*IBITS(s4,3,1)*gamma(2,3) + IBITS(s3,2,1)*(1-IBITS(s4,3,1))*gamma(2,2) + (1-IBITS(s3,2,1))*IBITS(s4,2,1)*gamma(2,2) + (1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*gamma(2,1) - IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) - IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) - (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) - (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) )  &   ! when upstream 1
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
        tb(j)=NORMCDF(( &
			-( utilsub(2) + IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(2,3) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(2,2) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(2,2) + (1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(2,1) + IBITS(s3,1,1)*IBITS(s4,2,1)*gamma(2,2) + IBITS(s3,1,1)*(1-IBITS(s4,2,1))*gamma(2,1) + (1-IBITS(s3,1,1))*IBITS(s4,1,1)*gamma(2,1) - IBITS(s3,0,1)*IBITS(s4,1,1)*alpha(2,2) - IBITS(s3,0,1)*(1-IBITS(s4,1,1))*alpha(2,1) - (1-IBITS(s3,0,1))*IBITS(s4,0,1)*alpha(2,1) )   &   ! when upstraem 0
        -DOT_PRODUCT(lchol(3,1:2),v(j,stmp,4:3:-1)))*temp,0.0d0,1.0d0)
    	END DO         ! Simulation loop

        DO j=1, ns
            IF (tb(j)-ta(j) > 0.0d0) THEN
            wg(j,stmp)=wg(j,stmp)*(tb(j)-ta(j))
       	    IF (ta(j)>=0.99999999d0) THEN 
!	            PRINT *,'Warning: a is too large.'
    		    v(j,stmp,2)=HUGE(v(j,stmp,2))
            ELSE IF (tb(j)<=0.00000001d0) THEN 
    !		    PRINT *,'Warning: b is too small.'
        	    v(j,stmp,2)=-HUGE(v(j,stmp,2))
   	        ELSE
          	    v(j,stmp,2)=INVNORMCDF((ta(j)*(1-draws(j,stmp,2))+tb(j)*draws(j,stmp,2)),0.0d0,1.0d0)
            ENDIF
            ELSE
                wg(j,stmp)=0
            ENDIF
        END DO

    ELSE

    	wg(:,stmp)=0.0d0
    	v(:,stmp,2)=0.0d0
        
    ENDIF


ELSE
    wg(:,s2*128+s3*16+s4)=0.0d0
ENDIF
END DO         ! Strategy loop
ELSE
    DO s4=0,2**4-1
    wg(:,s2*128+s3*16+s4)=0.0d0
    END DO
ENDIF
END DO         ! Strategy loop
ELSE
    DO s3=0,2**3-1
    DO s4=0,2**4-1
    wg(:,s2*128+s3*16+s4)=0.0d0
    END DO
    END DO
ENDIF
END DO         ! Strategy loop




! player 1

DO s2=0,2**2-1       ! Strategy loop
IF (mask(s2,2)==1) THEN
DO s3=0,2**3-1       ! Strategy loop
IF (mask(s3,3)==1) THEN
DO s4=0,2**4-1       ! Strategy loop
stmp=s2*128+s3*16+s4
IF (mask(s4,4)==1 .and. wg(1,stmp)>0.0d0) THEN

    utiltemp=utilsub(1) &
			+ IBITS(s2,1,1)*IBITS(s3,2,1)*alpha(1,3) &
			+ IBITS(s2,1,1)*(1-IBITS(s3,2,1))*IBITS(s4,2,1)*alpha(1,3) &
			+ IBITS(s2,1,1)*(1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*alpha(1,2) &
			+ (1-IBITS(s2,1,1))*IBITS(s3,1,1)*IBITS(s4,2,1)*alpha(1,3) &
			+ (1-IBITS(s2,1,1))*IBITS(s3,1,1)*(1-IBITS(s4,2,1))*alpha(1,2) &
			+ (1-IBITS(s2,1,1))*(1-IBITS(s3,1,1))*IBITS(s4,1,1)*alpha(1,2) &
			+ (1-IBITS(s2,1,1))*(1-IBITS(s3,1,1))*(1-IBITS(s4,1,1))*alpha(1,1) &
			+ IBITS(s2,1,1)*IBITS(s3,2,1)*IBITS(s4,3,1)*gamma(1,3) &
			+ IBITS(s2,1,1)*IBITS(s3,2,1)*(1-IBITS(s4,3,1))*gamma(1,2) &
			+ IBITS(s2,1,1)*(1-IBITS(s3,2,1))*IBITS(s4,2,1)*gamma(1,2) &
			+ IBITS(s2,1,1)*(1-IBITS(s3,2,1))*(1-IBITS(s4,2,1))*gamma(1,1) &
			+ (1-IBITS(s2,1,1))*IBITS(s3,1,1)*IBITS(s4,2,1)*gamma(1,2) &
			+ (1-IBITS(s2,1,1))*IBITS(s3,1,1)*(1-IBITS(s4,2,1))*gamma(1,1) &
			+ (1-IBITS(s2,1,1))*(1-IBITS(s3,1,1))*IBITS(s4,1,1)*gamma(1,1)
    IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==1 .and. IBITS(s4,2,1)==1) THEN
        utiltemp=utiltemp-alpha(1,3)
    ELSE IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==1 .and. IBITS(s4,2,1)==0) THEN
        utiltemp=utiltemp-alpha(1,2)
    ELSE IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==0 .and. IBITS(s4,1,1)==1) THEN
        utiltemp=utiltemp-alpha(1,2)
    ELSE IF (IBITS(s2,0,1)==0 .and. IBITS(s3,0,1)==1 .and. IBITS(s4,1,1)==1) THEN
        utiltemp=utiltemp-alpha(1,2)
    ELSE IF (IBITS(s2,0,1)==1 .and. IBITS(s3,1,1)==0 .and. IBITS(s4,1,1)==0) THEN
        utiltemp=utiltemp-alpha(1,1)
    ELSE IF (IBITS(s2,0,1)==0 .and. IBITS(s3,0,1)==1 .and. IBITS(s4,1,1)==0) THEN
        utiltemp=utiltemp-alpha(1,1)
    ELSE IF (IBITS(s2,0,1)==0 .and. IBITS(s3,0,1)==0 .and. IBITS(s4,0,1)==1) THEN
        utiltemp=utiltemp-alpha(1,1)
    ENDIF

	IF (mask(1,1)==1) THEN
		k = 8 + IBITS(s2,1,1)*4 + IBITS(s3,1+IBITS(s2,1,1),1)*2 + IBITS(s4,1+IBITS(s2,1,1)+IBITS(s3,1+IBITS(s2,1,1),1),1)
		temp=1.0d0/lchol(4,4)
		DO j=1, ns
			ta(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(4,1:3),v(j,stmp,4:2:-1)))*temp,0.0d0,1.0d0)
			IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,stmp)*(1.0-ta(j))
			IF (option==2) wgv(j,stmp,1) = INVNORMCDF((ta(j)*(1-draws(j,stmp,1))+draws(j,stmp,1)),0.0d0,1.0d0)
		END DO         ! Simulation loop
		IF (option==0) wg(:,stmp)=wg(:,stmp)*(1.0-ta)
		IF (option==2) wgv(:,stmp,0)=wg(:,stmp)*(1.0-ta)
	ENDIF
	IF (mask(0,1)==1) THEN
		k = IBITS(s2,0,1)*4 + IBITS(s3,IBITS(s2,0,1),1)*2 + IBITS(s4,IBITS(s2,0,1)+IBITS(s3,IBITS(s2,0,1),1),1)
		temp=1.0d0/lchol(4,4)
		DO j=1, ns
			tb(j)=NORMCDF((-utiltemp-DOT_PRODUCT(lchol(4,1:3),v(j,stmp,4:2:-1)))*temp,0.0d0,1.0d0)
			IF (option==1) wgv(k,1,1) = wgv(k,1,1) + wg(j,stmp) * tb(j)
			IF (option==2) wgv(j,stmp,1) = INVNORMCDF((tb(j)*draws(j,stmp,1)),0.0d0,1.0d0)
		END DO         ! Simulation loop
		IF (option==0) wg(:,stmp)=wg(:,stmp)*tb
		IF (option==2) wgv(:,stmp,0)=wg(:,stmp)*tb
    ENDIF

ENDIF
END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop
ENDIF
END DO         ! Strategy loop


IF (option==0) THEN
	wgv(1,1,1)=SUM(wg(:,:))	! wg(ns,s2*128+s3*16+s4)
	wgv=wgv/ns

	IF (wgv(1,1,1)<=0.0) THEN
	!	PRINT *,'Error found!'
	!	PRINT *,'GHK Value: ', ghkval
		wgv(1,1,1)=TINY(wgv(1,1,1))
	!	STOP
	ENDIF
END IF

IF (option==1) THEN
	wgv=wgv/ns
ENDIF

IF (option==2) THEN
	wgv(:,:,2)=v(:,:,2)
	wgv(:,:,3)=v(:,:,3)
	wgv(:,:,4)=v(:,:,4)
ENDIF


END FUNCTION GHK4





FUNCTION nearcounti(ent,n)
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: ent    ! family configuration
INTEGER, INTENT(IN) :: n      ! nchild
INTEGER :: i
INTEGER :: nearcounti

nearcounti=0
IF (n>=0) THEN
DO i=0, n-1
    nearcounti=nearcounti+IBITS(ent,i,1)
END DO
END IF

END FUNCTION nearcounti





FUNCTION lcholbase(m,param,n) RESULT(lchol)
USE Globvar
USE MATRIX
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: lchol(nchild(m),nchild(m)), omega(nchild(m),nchild(m))
INTEGER :: i

omega=0.0000d0
DO i = 1, nchild(m)
omega(i,i)=1.0000d0
END DO

SELECT CASE(nchild(m))
CASE(1)
omega(1,1)=1.0000d0
CASE(2)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

CASE(3)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,2,famindex(m)))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)-1))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

CASE(4)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,3,famindex(m)))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,2,famindex(m)))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,2,famindex(m)-1))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

omega(1,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)))
    IF (omega(1,4)<-0.8) omega(1,4)=-0.8d0-((DEXP(-omega(1,4)-0.8d0)/(1+DEXP(-omega(1,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,4)> 0.8) omega(1,4)= 0.8d0+((DEXP( omega(1,4)-0.8d0)/(1+DEXP( omega(1,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,1)=omega(1,4)

omega(2,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)-1))
    IF (omega(2,4)<-0.8) omega(2,4)=-0.8d0-((DEXP(-omega(2,4)-0.8d0)/(1+DEXP(-omega(2,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,4)> 0.8) omega(2,4)= 0.8d0+((DEXP( omega(2,4)-0.8d0)/(1+DEXP( omega(2,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,2)=omega(2,4)

omega(3,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xr(:,1,famindex(m)-2))
    IF (omega(3,4)<-0.8) omega(3,4)=-0.8d0-((DEXP(-omega(3,4)-0.8d0)/(1+DEXP(-omega(3,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(3,4)> 0.8) omega(3,4)= 0.8d0+((DEXP( omega(3,4)-0.8d0)/(1+DEXP( omega(3,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,3)=omega(3,4)

CASE DEFAULT

END SELECT

lchol=CHOL(omega)

END FUNCTION lcholbase




END MODULE GHK
