MODULE Globvar
IMPLICIT NONE
! This module declares the variables that are going to be used throughout the program.
! To avoid stack overflow error --> Stack Reserve: 50,000,000

INTEGER, PARAMETER :: nchildmax=4  ! max number of children
INTEGER :: nobs
INTEGER :: nfam
INTEGER :: ns              ! number of simulation draws
REAL(8) :: objsaved
INTEGER :: numeval
INTEGER :: ntheta
INTEGER :: nthetab     ! # of parameters in beta term
INTEGER :: nthetaa     ! # of parameters in alpha term
INTEGER :: nthetag     ! # of parameters in gamma term
INTEGER :: nthetar     ! # of parameters in rho term
INTEGER, ALLOCATABLE :: famid(:)   ! famid(nobs)
INTEGER, ALLOCATABLE :: nchild(:), famindex(:)   ! nchild(nfam), famindex(nfam)
REAL(8), ALLOCATABLE :: theta(:) ! theta(ntheta)
REAL(8), ALLOCATABLE :: xb(:,:)   ! xb(nthetab,nobs)
REAL(8), ALLOCATABLE :: xa(:,:)   ! xa(nthetaa,nobs)
REAL(8), ALLOCATABLE :: xg(:,:)   ! xg(nthetag,nobs)
INTEGER :: sa, sg		! actual size of xa and xg
REAL(8), ALLOCATABLE :: xr(:,:,:)   ! xr(nthetar,nchildmax,nobs) xr(:,i,:)
INTEGER, ALLOCATABLE :: y(:)     ! y(nobs)
INTEGER, ALLOCATABLE :: near(:)   ! near(nfam)
REAL(8), ALLOCATABLE :: astar(:)     ! astar(nobs) ... equilibrium decision rule (threshold value for the error term)
INTEGER, ALLOCATABLE :: predobs(:)  !nearpred(nobs)


REAL(8), ALLOCATABLE :: draws2seed(:,:)    ! nchild=2. draws(ns,2)
REAL(8), ALLOCATABLE :: draws3seed(:,:)    ! nchild=3. draws(ns,3)
REAL(8), ALLOCATABLE :: draws4seed(:,:)    ! nchild=4. draws(ns,4)
REAL(8), ALLOCATABLE :: draws2e(:,:)    ! nchild=2. draws(ns,2)	used in fixed point search to handle correlation
REAL(8), ALLOCATABLE :: draws3e(:,:)    ! nchild=3. draws(ns,3)	used in fixed point search to handle correlation
REAL(8), ALLOCATABLE :: draws4e(:,:)    ! nchild=4. draws(ns,4)	used in fixed point search to handle correlation

CHARACTER(10) :: path                          ! Specify the data directory. To be read from setting.prn

END MODULE Globvar



MODULE INIT
	USE Globvar
	IMPLICIT NONE

CONTAINS

	SUBROUTINE READDATA(nobs)
		USE Globvar
		IMPLICIT NONE
		INTEGER, INTENT(IN) :: nobs
		INTEGER :: i,j,k,m,io

        CHARACTER(60) :: readfile_x      ! File names to be read

readfile_x='data.raw'

xb=0.0d0; y=0.0d0
xb(1,:)=1.00000000d0; xa(1,:)=1.00000000d0
IF (nthetag>=1) xg(1,:)=1.00000000d0
xr(1,:,:)=1.00000000d0

print *,'Reading Data ... '

	OPEN(1,file=(readfile_x),status='old')

    IF (nthetaa>3 .and. nthetar>1 .and. nthetag>2) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xg(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>3 .and. nthetar>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i), xr(2:nthetar,1,i), xr(2:nthetar,2,i), xr(2:nthetar,3,i), xr(2:nthetar,4,i)
        IF (io<0) EXIT
        END DO
    ELSE IF (nthetaa>1) THEN
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i), xa(2:,i)
        IF (io<0) EXIT
        END DO
    ELSE
        DO i = 1, nobs
	    READ(1,fmt=*,iostat=io) y(i), famid(i), xb(2:,i)
        IF (io<0) EXIT
        END DO
    ENDIF
    CLOSE(1)


nchild=0
j=1
nchild(1)=1
DO i=2,nobs
	IF (famid(i)/=famid(i-1)) THEN
		j=j+1
	ENDIF
		nchild(j)=nchild(j)+1
ENDDO

famindex(1)=nchild(1)
DO i=2,nfam
	famindex(i)=famindex(i-1)+nchild(i)
ENDDO


DO m=1,nfam
near(m)=0
DO i=1,nchild(m)
near(m)=near(m)*2
near(m)=near(m)+y(famindex(m)-nchild(m)+i)
END DO
END DO



print *,'Reading Data ... Finished '

	END SUBROUTINE READDATA


END MODULE INIT

MODULE Object
	USE Globvar
	USE MATRIX
    USE PROBABILITY
	USE GHK
	IMPLICIT NONE

CONTAINS


FUNCTION lnmlobj(param,n,gradflag)	! This function computes the ML objective function
USE Globvar
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: gradflag
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: mlobj
REAL(8) :: lnmlobj
REAL(8) :: famprob(nfam)			! The family-level likelihood
REAL(8) :: atemp(nobs)
INTEGER :: i, j, k, m
REAL(4) :: tms
REAL(8) :: draws2(ns,2), draws2_2(ns,2), draws2_3(ns,2)
REAL(8) :: draws3(ns,3), draws3_2(ns,3), draws3_3(ns,3)
REAL(8) :: draws4(ns,4), draws4_2(ns,4), draws4_3(ns,4)

tms = SECNDS(0.0)

numeval=numeval+1

print ('(a16,I8)'),'mle evaluation: ',numeval


! This section is to elaborate the random draws (without this, every household uses the same set of 200 simulation draws)
! This is supposed to help convergence, though it doesn't affect the result much.
draws2=draws2seed+0.13
WHERE(draws2>1.0d0) draws2=draws2-1.0d0
draws3=draws3seed+0.13
WHERE(draws3>1.0d0) draws3=draws3-1.0d0
draws4=draws4seed+0.13
WHERE(draws4>1.0d0) draws4=draws4-1.0d0
    draws2=CSHIFT(draws2,200,1)
    draws3=CSHIFT(draws3,200,1)
    draws4=CSHIFT(draws4,200,1)
draws2_2=draws2+0.13
WHERE(draws2_2>1.0d0) draws2_2=draws2_2-1.0d0
draws3_2=draws3+0.13
WHERE(draws3_2>1.0d0) draws3_2=draws3_2-1.0d0
draws4_2=draws4+0.13
WHERE(draws4_2>1.0d0) draws4_2=draws4_2-1.0d0
    draws2_2=CSHIFT(draws2_2,200,1)
    draws3_2=CSHIFT(draws3_2,200,1)
    draws4_2=CSHIFT(draws4_2,200,1)
draws2_3=draws2_2+0.13
WHERE(draws2_3>1.0d0) draws2_3=draws2_3-1.0d0
draws3_3=draws3_2+0.13
WHERE(draws3_3>1.0d0) draws3_3=draws3_3-1.0d0
draws4_3=draws4_2+0.13
WHERE(draws4_3>1.0d0) draws4_3=draws4_3-1.0d0
    draws2_3=CSHIFT(draws2_3,200,1)
    draws3_3=CSHIFT(draws3_3,200,1)
    draws4_3=CSHIFT(draws4_3,200,1)

atemp=astar

!$OMP PARALLEL DO SCHEDULE(DYNAMIC)
DO m = 1, nfam

!WRITE(4,fmt=*) m, nchild(m)

SELECT CASE(nchild(m))
CASE(1)
	CALL LIKELIHOOD1(famprob(m),atemp(famindex(m)),m,param,n, &
		nthetab, nthetaa, xb(:,famindex(m)), xa(:,famindex(m)), sa, near(m))
CASE(2)
    IF (MOD(m,20)<5) THEN
	CALL LIKELIHOOD2(famprob(m),atemp(famindex(m)-1:famindex(m)),m,param,n, &
		ns,draws2seed,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<10) THEN
	CALL LIKELIHOOD2(famprob(m),atemp(famindex(m)-1:famindex(m)),m,param,n, &
		ns,draws2    ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<15) THEN
	CALL LIKELIHOOD2(famprob(m),atemp(famindex(m)-1:famindex(m)),m,param,n, &
		ns,draws2_2  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m))
	ELSE
	CALL LIKELIHOOD2(famprob(m),atemp(famindex(m)-1:famindex(m)),m,param,n, &
		ns,draws2_3  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-1:famindex(m)), xa(:,famindex(m)-1:famindex(m)), xg(:,famindex(m)-1:famindex(m)), sa, sg, near(m))
	ENDIF
CASE(3)
    IF (MOD(m,20)<5) THEN
	CALL LIKELIHOOD3(famprob(m),atemp(famindex(m)-2:famindex(m)),m,param,n, &
		ns,draws3seed,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<10) THEN
	CALL LIKELIHOOD3(famprob(m),atemp(famindex(m)-2:famindex(m)),m,param,n, &
		ns,draws3    ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<15) THEN
	CALL LIKELIHOOD3(famprob(m),atemp(famindex(m)-2:famindex(m)),m,param,n, &
		ns,draws3_2  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m))
	ELSE
	CALL LIKELIHOOD3(famprob(m),atemp(famindex(m)-2:famindex(m)),m,param,n, &
		ns,draws3_3  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-2:famindex(m)), xa(:,famindex(m)-2:famindex(m)), xg(:,famindex(m)-2:famindex(m)), sa, sg, near(m))
	ENDIF
CASE(4)

	IF (MOD(m,20)<5) THEN
	CALL LIKELIHOOD4(famprob(m),atemp(famindex(m)-3:famindex(m)),m,param,n, &
		ns,draws4seed,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<10) THEN
	CALL LIKELIHOOD4(famprob(m),atemp(famindex(m)-3:famindex(m)),m,param,n, &
		ns,draws4    ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m))
	ELSEIF (MOD(m,20)<15) THEN
	CALL LIKELIHOOD4(famprob(m),atemp(famindex(m)-3:famindex(m)),m,param,n, &
		ns,draws4_2  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m))
	ELSE
	CALL LIKELIHOOD4(famprob(m),atemp(famindex(m)-3:famindex(m)),m,param,n, &
		ns,draws4_3  ,lcholbase(m,param,n,nchild(m)),nthetab, nthetaa, nthetag, xb(:,famindex(m)-3:famindex(m)), xa(:,famindex(m)-3:famindex(m)), xg(:,famindex(m)-3:famindex(m)), sa, sg, near(m))
	ENDIF
CASE DEFAULT
!    PRINT *,'nchild(m): ', nchild(m)
!    PRINT *,'error'
!    PAUSE
END SELECT

END DO
!$OMP END PARALLEL DO

lnmlobj=(0-1)*SUM(famprob)

! Previously found optimal decision rule is used when there is an improvement in likelihood
IF (objsaved==0.0d0) THEN	! This applies only when the first time the likelihood function is evaluated.
	astar=atemp
	objsaved=-lnmlobj
ELSEIF (objsaved<-lnmlobj) THEN	! When a better likelihood is found
	astar=atemp
	objsaved=-lnmlobj
ENDIF


print *,'ln ML objective: ',-lnmlobj
!IF (gradflag==0) objsaved=mlobj

print *,'Time spent for Likelihood evaluation: ',SECNDS(tms)


END FUNCTION lnmlobj


SUBROUTINE LIKELIHOOD1(lhval,a,m,param,n,nb,na,xb,xa,sa,near0)
	! This function calculates the GHK simulator for each market*year
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: nb, na, sa
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: xb(nb),xa(sa)
INTEGER, INTENT(IN) :: near0
REAL(8), INTENT(OUT)  :: lhval		! The calculated likelihood value
REAL(8), INTENT(OUT)  :: a		! decision rule of the child. Choose near if e is larger than this value.
INTEGER :: i, j, k, l
REAL(8) :: utilsub, alpha

	utilsub=DOT_PRODUCT(param(1:nb),xb(:))
    IF(na>=3) THEN
    alpha  =EXP( DOT_PRODUCT(param(nb+1:nb+na-2),xa(:)) )
    ELSEIF(na==1) THEN
    alpha  =param(nb+na)
    ELSE
    alpha = 0.0d0
	ENDIF

	a = utilsub+alpha

	IF (IBITS(near0,0,1)==1) THEN
    lhval=LOG(1.0d0-NORMCDF(-a,0.0d0,1.0d0))
	ELSE
    lhval=LOG(NORMCDF(-a,0.0d0,1.0d0))
    ENDIF

END SUBROUTINE LIKELIHOOD1


SUBROUTINE LIKELIHOOD2(lhval,a,m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0)
	! This function calculates the log likelihood value for two child families.
	! It also computes the equilibrium strategy profile a* and saves it.
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,2)
REAL(8), INTENT(IN) :: lchol(2,2)
REAL(8), INTENT(IN) :: xb(nb,2),xa(sa,2),xg(sg,2)
INTEGER, INTENT(IN) :: near0
REAL(8), INTENT(OUT) :: lhval		! The calculated likelihood value
REAL(8), INTENT(INOUT) :: a(2)			! a(1) ... decision rule of the oldest child. Choose near if a(1)+e>0.
INTEGER :: i, j, k, l
REAL(8) :: prob, netutil, atemp(2)
REAL(8) :: th, lchol1(2,2), lchol2(2,2), v(2)
REAL(8) :: utilsub(2), alpha(2,2), gamma(2,2)		! 1 ... oldest child

!	PRINT *, 'LIKELIHOOD 2 - in!'

	! Preparation
	utilsub(1:2)=MATMUL(param(1:nb),xb(:,1:2))
	IF (na>=3) THEN
	alpha(1:2,1) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:2)) )
	alpha(1:2,2) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:2)) + param(nb+na-1) )
	ELSEIF (na==1) THEN
	alpha(1:2,1:2)=param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:2,1) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:2))
	gamma(1:2,2) =param(nb+na+ng)
	ELSEIF (ng==1) THEN
	gamma(1:2,1)=param(nb+na+ng)
	gamma(1:2,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

	! Setting initial decision rules, if nothing is previously saved.
	IF (SUM(a(:))==0.0d0) THEN
	a(2) = utilsub(2)+alpha(2,1)
	a(1) = utilsub(1)+alpha(1,1)
	ENDIF
	atemp=a

	lchol2=lchol
	lchol1=lcholbase(m,param,n,1)
		! Permulation for Cholesky decomposition is used to
		! compute the conditional distribution.

	DO k=1,20	! This DO loop searches the equilibrium strategy (i.e. participation threshold) as a fixed point.
		! Compute a new decision rule a* : 2nd child
		DO i=1,3	! This DO loop searches the optimal threshold for the 2nd child given a().
					! Numerical iteration is necessary to deal with the correlation in the error terms.
			netutil=0.0d0
			! case: s1=1 (the first child lives near)
			v(2)=-atemp(2)/lchol2(1,1)
				! indifferent e2=-atemp(2)=v2*lchol2(1,1)
			th=(-a(1)-lchol2(2,1)*v(2))/lchol2(2,2)
				! The above lines computes the conditional distribution of e2 based on atemp(2)
			IF (th>-999.0 .and. th<999.0) THEN	! When th becomes extremely small or large, use the formula for no correlation.
				prob = 1.0d0 - NORMCDF(th,0.0d0,1.0d0)
			ELSE
				prob=(1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob*(utilsub(2) + alpha(2,2) - alpha(2,1) +gamma(2,1))
				! Net util gain = utilsub(2)+alpha(2,1)*(1.0d0+alpha(2,2))+gamma(2,1)-alpha(2,1)
			! case: s1=0
			netutil=netutil+(1.0d0-prob)*(utilsub(2) + alpha(2,1))
				! Net util gain = utilsub(2)+alpha(2,1)
			atemp(2)=netutil
		ENDDO

		! Solving for the equilibrium decision rule a* : 1st child
		DO i=1,3
			netutil=0.0d0
			! case: s2=1
			v(1)=-atemp(1)/lchol1(1,1)
				! indifferent e1=-atemp(1)=v1*lchol1(1,1)
			th=(-a(2)-lchol1(2,1)*v(1))/lchol1(2,2)
			IF (th>-999.0 .and. th<999.0) THEN
				prob = 1.0d0 - NORMCDF(th,0.0d0,1.0d0)
			ELSE
				prob=(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob*(utilsub(1) + alpha(1,2) - alpha(1,1) +gamma(1,1))
				! Net util gain = utilsub(1)+alpha(1,1)*(1.0d0+alpha(1,2))+gamma(1,1)-alpha(1,1)
			! case: s2=0
			netutil=netutil+(1.0d0-prob)*(utilsub(1)+alpha(1,1))
				! Net util gain = utilsub(1)+alpha(1,1)
			atemp(1)=netutil
		ENDDO
		a=atemp 
	ENDDO


	! Computing the likelihood given equilibrium decision rule a*
	lhval=LOG(GHK2(param,n,ns,draws,lchol,a,near0))

!	PRINT *, 'LIKELIHOOD 2 - out!'

END SUBROUTINE LIKELIHOOD2


SUBROUTINE LIKELIHOOD3(lhval,a,m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0)
	! This function calculates the log likelihood value for three child families.
	! It also computes the equilibrium strategy profile a* and saves it.
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,3)
REAL(8), INTENT(IN) :: lchol(3,3)
REAL(8), INTENT(IN) :: xb(nb,3),xa(sa,3),xg(sg,3)
INTEGER, INTENT(IN) :: near0
REAL(8), INTENT(OUT) :: lhval		! The calculated likelihood value
REAL(8), INTENT(INOUT) :: a(3)			! a(1) ... decision rule of the oldest child. Choose near if e is larger than this.
INTEGER :: i, j, k, l
REAL(8) :: prob(3), netutil, atemp(3)
REAL(8) :: th1, th2, th(ns), lchol1(3,3), lchol2(3,3), lchol3(3,3), v(ns,3), wg, temp
REAL(8) :: utilsub(3), alpha(3,3), gamma(3,2)		! 1 ... oldest child

!	PRINT *, 'LIKELIHOOD 3 - in!'

	utilsub(1:3) = MATMUL(param(1:nb),xb(:,1:3))
	IF (na>=3) THEN
	alpha(1:3,1) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) )
	alpha(1:3,2) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) + param(nb+na-1) )
	alpha(1:3,3) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:3)) + param(nb+na-1) + param(nb+na) )
	ELSEIF (na==1) THEN
	alpha(1:3,1:3)=param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:3,1) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:3))
	gamma(1:3,2) =param(nb+na+ng)
	ELSEIF (ng==1) THEN
	gamma(1:3,1)=param(nb+na+ng)
	gamma(1:3,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

	! Setting initial decision rules
	
	IF (SUM(a(:))==0.0d0) THEN
	a(3) = utilsub(3)+alpha(3,1)
	a(2) = utilsub(2)+alpha(2,1)
	a(1) = utilsub(1)+alpha(1,1)
	ENDIF
	atemp=a

	lchol3=lchol
	lchol2=lcholbase(m,param,n,2)
	lchol1=lcholbase(m,param,n,1)
		! Permulation for Cholesky decomposition is used to
		! compute the conditional distribution.

	DO k=1,20	! This DO loop searches the equilibrium strategy (i.e. participation threshold) as a fixed point.
		! Compute a new decision rule a* : 3rd child
		temp=1.0d0/lchol3(3,3)	! This line is to speed up computation.
		DO i=1,3	! This DO loop searches the optimal threshold for the 2nd child given a().
					! Numerical iteration is necessary to deal with the correlation in the error terms.
			netutil=0.0d0
			v(1,3)=-atemp(3)/lchol3(1,1)
			th2=(-a(2)-lchol3(2,1)*v(1,3))/lchol3(2,2)
			wg=NORMCDF(th2,0.0d0,1.0d0)
			! case: s1+s2=2
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg*(1.0d0-draws3e(j,2))+draws3e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(3,1)*v(1,3)-lchol3(3,2)*v(j,2))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN
				prob(3) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(3) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(3)*(utilsub(3)+alpha(3,3)-alpha(3,2)+gamma(3,1)*(1.0d0+gamma(3,2)))
				! Net util gain = utilsub(3)+alpha(3,1)*(1.0d0+alpha(3,2)+alpha(3,3))+gamma(3,1)*(1.0d0+gamma(3,2))-alpha(3,1)*(1.0d0+alpha(3,2))
			! case: s1+s2=1
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN	! only 2nd child near
				prob(1) = NORMCDF(th1,0.0d0,1.0d0)  *(1.0d0-NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))
			ENDIF

			DO j=1, ns
				v(j,2)=INVNORMCDF(wg*draws3e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(3,1)*v(1,3)-lchol3(3,2)*v(j,2))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN	! only 1st child near
				prob(2) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0)
			ELSE
				prob(2) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(2),0.0d0,1.0d0)
			ENDIF

			netutil=netutil+(prob(2)+prob(1))*(utilsub(3)+alpha(3,2)-alpha(3,1)+gamma(3,1))
				! Net util gain = utilsub(3)+alpha(3,1)*(1.0d0+alpha(3,2))+gamma(3,1)-alpha(3,1)
			! case: s1+s2=0
			netutil=netutil+(1.0d0-SUM(prob(1:3)))*(utilsub(3)+alpha(3,1))
				! Net util gain = utilsub(3)+alpha(3,1)
			atemp(3)=netutil
		ENDDO
		! Compute a new decision rule a* : 2nd child
		temp=1.0d0/lchol2(3,3)
		DO i=1,3
			netutil=0.0d0
			v(1,2)=-atemp(2)/lchol2(1,1)
			th2=(-a(3)-lchol2(2,1)*v(1,2))/lchol2(2,2)
			wg=NORMCDF(th2,0.0d0,1.0d0)
			! case: s1+s3=2
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg*(1.0d0-draws3e(j,3))+draws3e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(3,1)*v(1,2)-lchol2(3,2)*v(j,3))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN
				prob(3) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(3)=(1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(3)*(utilsub(2)+alpha(2,3)-alpha(2,2)+gamma(2,1)*(1.0d0+gamma(2,2)))
			! case: s1+s3=1
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF

			DO j=1, ns
				v(j,3)=INVNORMCDF(wg*draws3e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(3,1)*v(1,2)-lchol2(3,2)*v(j,3))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN	! only 1st child near
				prob(2) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0)
			ELSE
				prob(2) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(2)+prob(1))*(utilsub(2)+alpha(2,2)-alpha(2,1)+gamma(2,1))
			! case: s1+s3=0
			netutil=netutil+(1.0d0-SUM(prob(1:3)))*(utilsub(2)+alpha(2,1))
			atemp(2)=netutil
		ENDDO
	
		! Solving for the equilibrium decision rule a* : 1st child
		temp=1.0d0/lchol1(3,3)
		DO i=1,3
			netutil=0.0d0
			v(1,1)=-atemp(1)/lchol1(1,1)
			th2=(-a(3)-lchol1(2,1)*v(1,1))/lchol1(2,2)
			wg=NORMCDF(th2,0.0d0,1.0d0)
			! case: s2+s3=2
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg*(1.0d0-draws3e(j,3))+draws3e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(3,1)*v(1,1)-lchol1(3,2)*v(j,3))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN
				prob(3) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(3)=(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(3)*(utilsub(1)+alpha(1,3)-alpha(1,2)+gamma(1,1)*(1.0d0+gamma(1,2)))
			! case: s2+s3=1
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg*draws3e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(3,1)*v(1,1)-lchol1(3,2)*v(j,3))*temp
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0) THEN	! only 1st child near
				prob(2) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0)
			ELSE
				prob(2) = (1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(2)+prob(1))*(utilsub(1)+alpha(1,2)-alpha(1,1)+gamma(1,1))
			! case: s2+s3=0
			netutil=netutil+(1.0d0-SUM(prob(1:3)))*(utilsub(1)+alpha(1,1))
			atemp(1)=netutil
		ENDDO
		a=atemp 
	ENDDO
	
	
	! Computing the likelihood given equilibrium decision rule a*
	
	lhval=LOG(GHK3(param,n,ns,draws,lchol,a,near0))

!	PRINT *, 'LIKELIHOOD 3 - out!'

END SUBROUTINE LIKELIHOOD3


SUBROUTINE LIKELIHOOD4(lhval,a,m,param,n,ns,draws,lchol,nb,na,ng,xb,xa,xg,sa,sg,near0)
	! This function calculates the log likelihood value for four child families.
	! It also computes the equilibrium strategy profile a* and saves it.
USE PROBABILITY
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n, ns
INTEGER, INTENT(IN) :: nb, na, ng, sa, sg	! # of parameters in beta term, alpha term, and gamma term
REAL(8), INTENT(IN) :: param(n)
REAL(8), INTENT(IN) :: draws(ns,4)
REAL(8), INTENT(IN) :: lchol(4,4)
REAL(8), INTENT(IN) :: xb(nb,4),xa(sa,4),xg(sg,4)
INTEGER, INTENT(IN) :: near0
REAL(8), INTENT(OUT) :: lhval		! The calculated likelihood value
REAL(8), INTENT(INOUT) :: a(4)			! a(1) ... decision rule of the oldest child. Choose near if e is larger than this.
INTEGER :: i, j, k, l
REAL(8) :: prob(7), netutil, atemp(4)
REAL(8) :: th1, th2, th3, th4, th(ns), lchol1(4,4), lchol2(4,4), lchol3(4,4), lchol4(4,4), v(ns,3), wg0, wg(ns), temp1(2:4), temp2(2:4), temp3(2:4), temp4(2:4)
REAL(8) :: utilsub(4), alpha(4,4), gamma(4,2)		! 1 ... oldest child

!	PRINT *, 'LIKELIHOOD 4 - in!'

	utilsub(1:4) = MATMUL(param(1:nb),xb(:,1:4))
	IF (na>=3) THEN
	alpha(1:4,1) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) )
	alpha(1:4,2) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) + param(nb+na-1) )
	alpha(1:4,3) = EXP( MATMUL(param(nb+1:nb+na-2),xa(:,1:4)) + param(nb+na-1) + param(nb+na) )
	ELSEIF (na==1) THEN
	alpha(1:4,1:3)=param(nb+na)
	ELSE
	alpha=0.0d0
	ENDIF
	IF (ng>=2) THEN
	gamma(1:4,1) =MATMUL(param(nb+na+1:nb+na+ng-1),xg(:,1:4))
	gamma(1:4,2) =param(nb+na+ng)
	ELSEIF (ng==1) THEN
	gamma(1:4,1)=param(nb+na+ng)
	gamma(1:4,2)=0.0d0
	ELSE
	gamma=0.0d0
	ENDIF

	! Setting initial decision rules
	
	IF (SUM(a(:))==0.0d0) THEN
	a(4) = utilsub(4)+alpha(4,1)
	a(3) = utilsub(3)+alpha(3,1)
	a(2) = utilsub(2)+alpha(2,1)
	a(1) = utilsub(1)+alpha(1,1)
	ENDIF
	atemp=a

	lchol4=lchol
	lchol3=lcholbase(m,param,n,3)
	lchol2=lcholbase(m,param,n,2)
	lchol1=lcholbase(m,param,n,1)
		! Permulation for Cholesky decomposition is used to
		! compute the conditional distribution.

	temp4(2)=1.0d0/lchol4(2,2)
	temp4(3)=1.0d0/lchol4(3,3)
	temp4(4)=1.0d0/lchol4(4,4)
		! These lines are to speed up computation.

	DO k=1,20	! This DO loop searches the equilibrium strategy (i.e. participation threshold) as a fixed point.
		! Compute a new decision rule a* : 4th child
		DO i=1,3	! This DO loop searches the optimal threshold for the 2nd child given a().
					! Numerical iteration is necessary to deal with the correlation in the error terms.
			netutil=0.0d0
			v(1,4)=-atemp(4)/lchol4(1,1)
			th3=(-a(3)-lchol4(2,1)*v(1,4))*temp4(2)
			wg0=NORMCDF(th3,0.0d0,1.0d0)
			! case: s1+s2+s3=3
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg0*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol4(3,1)*v(1,4)-lchol4(3,2)*v(j,3))*temp4(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,2))+draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol4(4,1)*v(1,4)-lchol4(4,2)*v(j,3)-lchol4(4,3)*v(j,2))*temp4(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(7) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(7) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(7)*(utilsub(4)+gamma(4,1)*(1.0d0+gamma(4,2)*2.0d0))
				! Net util gain = utilsub(4)+alpha(4,1)*(1.0d0+alpha(4,2)+alpha(4,3))+gamma(4,1)*(1.0d0+gamma(4,2)*2.0d0)-alpha(4,1)*(1.0d0+alpha(4,2)+alpha(4,3))

			! case: s1+s2+s3=2
				! s1=0, s2=s3=1
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(3) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(3) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
				! s2=0, s1=s3=1
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol4(4,1)*v(1,4)-lchol4(4,2)*v(j,3)-lchol4(4,3)*v(j,2))*temp4(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(5) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(5) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
				! s1=s2=1, s3=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg0*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol4(3,1)*v(1,4)-lchol4(3,2)*v(j,3))*temp4(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,2))+draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol4(4,1)*v(1,4)-lchol4(4,2)*v(j,3)-lchol4(4,3)*v(j,2))*temp4(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(6) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(6) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0)) * NORMCDF(-a(3),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(6)+prob(5)+prob(3))*(utilsub(4)+alpha(4,3)-alpha(4,2)+gamma(4,1)*(1.0d0+gamma(4,2)))
				! Net util gain = utilsub(4)+alpha(4,1)*(1.0d0+alpha(4,2)+alpha(4,3))+gamma(4,1)*(1.0d0+gamma(4,2))-alpha(4,1)*(1.0d0+alpha(4,2))
			! case: s1+s2+s3=1
				! s2=1, s1=s3=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(2) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(2) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0)) * NORMCDF(-a(3),0.0d0,1.0d0)
			ENDIF
				! s1=1, s2=s3=0
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol4(4,1)*v(1,4)-lchol4(4,2)*v(j,3)-lchol4(4,3)*v(j,2))*temp4(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(4) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(4) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(2),0.0d0,1.0d0) * NORMCDF(-a(3),0.0d0,1.0d0)
			ENDIF
				! s3=1, s1=s2=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg0*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol4(3,1)*v(1,4)-lchol4(3,2)*v(j,3))*temp4(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol4(4,1)*v(1,4)-lchol4(4,2)*v(j,3)-lchol4(4,3)*v(j,2))*temp4(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(1),0.0d0,1.0d0)*NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+(prob(4)+prob(2)+prob(1))*(utilsub(4)+alpha(4,2)-alpha(4,1)+gamma(4,1))
				! Net util gain = utilsub(4)+alpha(4,1)*(1.0d0+alpha(4,2))+gamma(4,1)-alpha(4,1)
			! case: s1+s2+s3=0
			netutil=netutil+(1.0d0-SUM(prob(1:7)))*(utilsub(4)+alpha(4,1))
				! Net util gain = utilsub(4)+alpha(4,1)
			atemp(4)=netutil
		ENDDO
		! Compute a new decision rule a* : 3rd child
		DO i=1,3
			netutil=0.0d0
			v(1,3)=-atemp(3)/lchol3(1,1)
			th3=(-a(4)-lchol3(2,1)*v(1,3))*temp3(2)
			wg0=NORMCDF(th3,0.0d0,1.0d0)
			! case: s1+s2+s4=3
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol3(3,1)*v(1,3)-lchol3(3,2)*v(j,4))*temp3(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,2))+draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(4,1)*v(1,3)-lchol3(4,2)*v(j,4)-lchol3(4,3)*v(j,2))*temp3(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(7) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(7) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(7)*(utilsub(3)+gamma(3,1)*(1.0d0+gamma(3,2)*2.0d0))
			! case: s1+s2+s4=2
				! s2=s4=1, s1=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(3) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(3) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s1=s4=1, s2=0
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(4,1)*v(1,3)-lchol3(4,2)*v(j,4)-lchol3(4,3)*v(j,2))*temp3(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(5) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(5) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s1=s2=1, s4=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol3(3,1)*v(1,3)-lchol3(3,2)*v(j,4))*temp3(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,2))+draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(4,1)*v(1,3)-lchol3(4,2)*v(j,4)-lchol3(4,3)*v(j,2))*temp3(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(6) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(6) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(6)+prob(5)+prob(3))*(utilsub(3)+alpha(3,3)-alpha(3,2)+gamma(3,1)*(1.0d0+gamma(3,2)))
			! case: s1+s2+s4=1
				! s2=1, s1=s4=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(2) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(2) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s1=1, s2=s4=0
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(4,1)*v(1,3)-lchol3(4,2)*v(j,4)-lchol3(4,3)*v(j,2))*temp3(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(4) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(4) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(2),0.0d0,1.0d0)* NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s4=1, s1=s2=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol3(3,1)*v(1,3)-lchol3(3,2)*v(j,4))*temp3(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,2)=INVNORMCDF(wg(j)*draws4e(j,2),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol3(4,1)*v(1,3)-lchol3(4,2)*v(j,4)-lchol3(4,3)*v(j,2))*temp3(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(1),0.0d0,1.0d0)*NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+(prob(4)+prob(2)+prob(1))*(utilsub(3)+alpha(3,2)-alpha(3,1)+gamma(3,1))
			! case: s1+s2+s4=0
			netutil=netutil+(1.0d0-SUM(prob(1:7)))*(utilsub(3)+alpha(3,1))
			atemp(3)=netutil
		ENDDO
		! Compute a new decision rule a* : 2nd child
		DO i=1,3
			netutil=0.0d0
			v(1,2)=-atemp(2)/lchol2(1,1)
			th3=(-a(4)-lchol2(2,1)*v(1,2))*temp2(2)
			wg0=NORMCDF(th3,0.0d0,1.0d0)
			! case: s1+s3+s4=3
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol2(3,1)*v(1,2)-lchol2(3,2)*v(j,4))*temp2(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(4,1)*v(1,2)-lchol2(4,2)*v(j,4)-lchol2(4,3)*v(j,3))*temp2(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(7) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(7) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(7)*(utilsub(2)+gamma(2,1)*(1.0d0+gamma(2,2)*2.0d0))
			! case: s1+s3+s4=2
				! s3=s4=1, s1=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(3) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(3) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s1=s4=1, s3=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(4,1)*v(1,2)-lchol2(4,2)*v(j,4)-lchol2(4,3)*v(j,3))*temp2(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(5) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(5) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s1=s3=1, s4=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol2(3,1)*v(1,2)-lchol2(3,2)*v(j,4))*temp2(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(4,1)*v(1,2)-lchol2(4,2)*v(j,4)-lchol2(4,3)*v(j,3))*temp2(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(6) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(6) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(6)+prob(5)+prob(3))*(utilsub(2)+alpha(2,3)-alpha(2,2)+gamma(2,1)*(1.0d0+gamma(2,2)))
			! case: s1+s3+s4=1
				! s3=1, s1=s4=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(2) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(2) = NORMCDF(-a(1),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s1=1, s3=s4=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(4,1)*v(1,2)-lchol2(4,2)*v(j,4)-lchol2(4,3)*v(j,3))*temp2(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(4) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(4) = (1.0d0-NORMCDF(-a(1),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s4=1, s1=s3=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol2(3,1)*v(1,2)-lchol2(3,2)*v(j,4))*temp2(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(1)-lchol2(4,1)*v(1,2)-lchol2(4,2)*v(j,4)-lchol2(4,3)*v(j,3))*temp2(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(1),0.0d0,1.0d0)*NORMCDF(-a(3),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+(prob(4)+prob(2)+prob(1))*(utilsub(2)+alpha(2,2)-alpha(2,1)+gamma(2,1))
			! case: s1+s3+s4=0
			netutil=netutil+(1.0d0-SUM(prob(1:7)))*(utilsub(2)+alpha(2,1))
			atemp(2)=netutil
		ENDDO
		! Compute a new decision rule a* : 1st child
		DO i=1,3
			netutil=0.0d0
			v(1,1)=-atemp(1)/lchol1(1,1)
			th3=(-a(4)-lchol1(2,1)*v(1,1))*temp1(2)
			wg0=NORMCDF(th3,0.0d0,1.0d0)
			! case: s2+s3+s4=3
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol1(3,1)*v(1,1)-lchol1(3,2)*v(j,4))*temp1(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(4,1)*v(1,1)-lchol1(4,2)*v(j,4)-lchol1(4,3)*v(j,3))*temp1(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(7) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(7) = (1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+prob(7)*(utilsub(1)+gamma(1,1)*(1.0d0+gamma(1,2)*2.0d0))
			! case: s2+s3+s4=2
				! s3=s4=1, s2=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(3) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(3) = NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s2=s4=1, s3=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(4,1)*v(1,1)-lchol1(4,2)*v(j,4)-lchol1(4,3)*v(j,3))*temp1(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(5) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(5) = (1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
				! s2=s3=1, s4=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol1(3,1)*v(1,1)-lchol1(3,2)*v(j,4))*temp1(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*(1.0d0-draws4e(j,3))+draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(4,1)*v(1,1)-lchol1(4,2)*v(j,4)-lchol1(4,3)*v(j,3))*temp1(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(6) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(6) = (1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))* NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
			netutil=netutil+(prob(6)+prob(5)+prob(3))*(utilsub(1)+alpha(1,3)-alpha(1,2)+gamma(1,1)*(1.0d0+gamma(1,2)))
			! case: s2+s3+s4=1
				! s3=1, s2=s4=0
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(2) = NORMCDF(th1,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th2,0.0d0,1.0d0)) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(2) = NORMCDF(-a(2),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(3),0.0d0,1.0d0))* NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s2=1, s3=s4=0
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(4,1)*v(1,1)-lchol1(4,2)*v(j,4)-lchol1(4,3)*v(j,3))*temp1(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(4) = (1.0d0 - NORMCDF(th1,0.0d0,1.0d0)) * NORMCDF(th2,0.0d0,1.0d0) * NORMCDF(th3,0.0d0,1.0d0)
			ELSE
				prob(4) = (1.0d0-NORMCDF(-a(2),0.0d0,1.0d0))*NORMCDF(-a(3),0.0d0,1.0d0)*NORMCDF(-a(4),0.0d0,1.0d0)
			ENDIF
				! s4=1, s2=s3=0
			DO j=1, ns
				v(j,4)=INVNORMCDF(wg0*(1.0d0-draws4e(j,4))+draws4e(j,4),0.0d0,1.0d0)
				th(j)=(-a(3)-lchol1(3,1)*v(1,1)-lchol1(3,2)*v(j,4))*temp1(3)
				wg(j)=NORMCDF(th(j),0.0d0,1.0d0)
			ENDDO
			th2=SUM(th(:))/ns
			DO j=1, ns
				v(j,3)=INVNORMCDF(wg(j)*draws4e(j,3),0.0d0,1.0d0)
				th(j)=(-a(2)-lchol1(4,1)*v(1,1)-lchol1(4,2)*v(j,4)-lchol1(4,3)*v(j,3))*temp1(4)
			ENDDO
			th1=SUM(th(:))/ns
			IF (th1>-999.0 .and. th1<999.0 .and. th2>-999.0 .and. th2<999.0 .and. th3>-999.0 .and. th3<999.0) THEN
				prob(1) = NORMCDF(th1,0.0d0,1.0d0) * NORMCDF(th2,0.0d0,1.0d0) * (1.0d0 - NORMCDF(th3,0.0d0,1.0d0))
			ELSE
				prob(1) = NORMCDF(-a(2),0.0d0,1.0d0)*NORMCDF(-a(3),0.0d0,1.0d0)*(1.0d0-NORMCDF(-a(4),0.0d0,1.0d0))
			ENDIF
			netutil=netutil+(prob(4)+prob(2)+prob(1))*(utilsub(1)+alpha(1,2)-alpha(1,1)+gamma(1,1))
			! case: s2+s3+s4=0
			netutil=netutil+(1.0d0-SUM(prob(1:7)))*(utilsub(1)+alpha(1,1))
			atemp(1)=netutil
		ENDDO
		a=atemp 
	ENDDO
	
	! Computing the likelihood given equilibrium decision rule a*
	
	lhval=LOG(GHK4(param,n,ns,draws,lchol,a,near0))

!	PRINT *, 'LIKELIHOOD 4 - out!'

END SUBROUTINE LIKELIHOOD4

FUNCTION lcholbase(m,param,n,option) RESULT(lchol)
! This function first constructs the VCOV matrix, and then 
! generates the Cholesky decomposition matrix.
! The default option (option==nchild(m)) assumes the first simulation
! draw for the youngest child. If a value of "option" < nchild(m) is assigned,
! say, i=2, the first simulation draw is made for the second child.
USE Globvar
USE MATRIX
IMPLICIT NONE
INTEGER, INTENT(IN) :: m			! family id
INTEGER, INTENT(IN) :: n 
INTEGER, INTENT(IN) :: option		! option to construct permutation.
	! usually set nchild(m) so that the youngest child hits the first draw.
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: xrtemp(SIZE(xr,1),nchild(m),nchild(m))
			! xrtemp(:,j,i) ... relational variables of child i's in relation to child j
REAL(8) :: lchol(nchild(m),nchild(m)), omega(nchild(m),nchild(m))
INTEGER :: i

IF (option<1 .or. option>nchild(m)) THEN
	print *,'Option: this value is not allowed.'
	STOP
ENDIF

IF (option==nchild(m)) THEN
	xrtemp=xr(:,1:nchild(m),famindex(m)-nchild(m)+1:famindex(m))
ELSE
	IF (nchild(m)==2 .and. option==1) THEN
		xrtemp(:,2,2)=xr(:,1,1)
		xrtemp(:,1,2)=xr(:,2,1)
		xrtemp(:,2,1)=xr(:,1,2)
		xrtemp(:,1,1)=xr(:,2,2)
	ELSEIF (nchild(m)==3 .and. option==1) THEN
		xrtemp(:,3,3)=xr(:,1,1)
		xrtemp(:,2,3)=xr(:,3,1)
		xrtemp(:,1,3)=xr(:,2,1)
		xrtemp(:,3,1)=xr(:,1,2)
		xrtemp(:,2,1)=xr(:,3,2)
		xrtemp(:,1,1)=xr(:,2,2)
		xrtemp(:,3,2)=xr(:,1,3)
		xrtemp(:,2,2)=xr(:,3,3)
		xrtemp(:,1,2)=xr(:,2,3)
	ELSEIF (nchild(m)==3 .and. option==2) THEN
		xrtemp(:,3,3)=xr(:,2,2)
		xrtemp(:,2,3)=xr(:,3,2)
		xrtemp(:,1,3)=xr(:,1,2)
		xrtemp(:,3,1)=xr(:,2,1)
		xrtemp(:,2,1)=xr(:,3,1)
		xrtemp(:,1,1)=xr(:,1,1)
		xrtemp(:,3,2)=xr(:,2,3)
		xrtemp(:,2,2)=xr(:,3,3)
		xrtemp(:,1,2)=xr(:,1,3)
	ELSEIF (nchild(m)==4 .and. option==1) THEN
		xrtemp(:,4,4)=xr(:,1,1)
		xrtemp(:,3,4)=xr(:,4,1)
		xrtemp(:,2,4)=xr(:,3,1)
		xrtemp(:,1,4)=xr(:,2,1)
		xrtemp(:,4,3)=xr(:,1,4)
		xrtemp(:,3,3)=xr(:,4,4)
		xrtemp(:,2,3)=xr(:,3,4)
		xrtemp(:,1,3)=xr(:,2,4)
		xrtemp(:,4,2)=xr(:,1,3)
		xrtemp(:,3,2)=xr(:,4,3)
		xrtemp(:,2,2)=xr(:,3,3)
		xrtemp(:,1,2)=xr(:,2,3)
		xrtemp(:,4,1)=xr(:,1,2)
		xrtemp(:,3,1)=xr(:,4,2)
		xrtemp(:,2,1)=xr(:,3,2)
		xrtemp(:,1,1)=xr(:,2,2)
	ELSEIF (nchild(m)==4 .and. option==2) THEN
		xrtemp(:,4,4)=xr(:,2,2)
		xrtemp(:,3,4)=xr(:,4,2)
		xrtemp(:,2,4)=xr(:,3,2)
		xrtemp(:,1,4)=xr(:,1,2)
		xrtemp(:,4,3)=xr(:,2,4)
		xrtemp(:,3,3)=xr(:,4,4)
		xrtemp(:,2,3)=xr(:,3,4)
		xrtemp(:,1,3)=xr(:,1,4)
		xrtemp(:,4,2)=xr(:,2,3)
		xrtemp(:,3,2)=xr(:,4,3)
		xrtemp(:,2,2)=xr(:,3,3)
		xrtemp(:,1,2)=xr(:,1,3)
		xrtemp(:,4,1)=xr(:,2,1)
		xrtemp(:,3,1)=xr(:,4,1)
		xrtemp(:,2,1)=xr(:,3,1)
		xrtemp(:,1,1)=xr(:,1,1)
	ELSEIF (nchild(m)==4 .and. option==3) THEN
		xrtemp(:,4,4)=xr(:,3,3)
		xrtemp(:,3,4)=xr(:,4,3)
		xrtemp(:,2,4)=xr(:,2,3)
		xrtemp(:,1,4)=xr(:,1,3)
		xrtemp(:,4,3)=xr(:,3,4)
		xrtemp(:,3,3)=xr(:,4,4)
		xrtemp(:,2,3)=xr(:,2,4)
		xrtemp(:,1,3)=xr(:,1,4)
		xrtemp(:,4,2)=xr(:,3,2)
		xrtemp(:,3,2)=xr(:,4,2)
		xrtemp(:,2,2)=xr(:,2,2)
		xrtemp(:,1,2)=xr(:,1,2)
		xrtemp(:,4,1)=xr(:,3,1)
		xrtemp(:,3,1)=xr(:,4,1)
		xrtemp(:,2,1)=xr(:,2,1)
		xrtemp(:,1,1)=xr(:,1,1)
	ELSE
		print *,'lcholbase error'
		stop
	ENDIF
ENDIF

omega=0.0000d0
DO i = 1, nchild(m)
omega(i,i)=1.0000d0
END DO

SELECT CASE(nchild(m))
CASE(1)
omega(1,1)=1.0000d0
CASE(2)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,2))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

CASE(3)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,2,3))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,3))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,2))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

CASE(4)

omega(1,2)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,3,4))
    ! The following lines helps to contain omega values between -0.9 and 0.9 and get convergence
    IF (omega(1,2)<-0.8) omega(1,2)=-0.8d0-((DEXP(-omega(1,2)-0.8d0)/(1+DEXP(-omega(1,2)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,2)> 0.8) omega(1,2)= 0.8d0+((DEXP( omega(1,2)-0.8d0)/(1+DEXP( omega(1,2)-0.8d0)))-0.5d0)*0.2d0
omega(2,1)=omega(1,2)

omega(1,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,2,4))
    IF (omega(1,3)<-0.8) omega(1,3)=-0.8d0-((DEXP(-omega(1,3)-0.8d0)/(1+DEXP(-omega(1,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,3)> 0.8) omega(1,3)= 0.8d0+((DEXP( omega(1,3)-0.8d0)/(1+DEXP( omega(1,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,1)=omega(1,3)

omega(2,3)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,2,3))
    IF (omega(2,3)<-0.8) omega(2,3)=-0.8d0-((DEXP(-omega(2,3)-0.8d0)/(1+DEXP(-omega(2,3)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,3)> 0.8) omega(2,3)= 0.8d0+((DEXP( omega(2,3)-0.8d0)/(1+DEXP( omega(2,3)-0.8d0)))-0.5d0)*0.2d0
omega(3,2)=omega(2,3)

omega(1,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,4))
    IF (omega(1,4)<-0.8) omega(1,4)=-0.8d0-((DEXP(-omega(1,4)-0.8d0)/(1+DEXP(-omega(1,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(1,4)> 0.8) omega(1,4)= 0.8d0+((DEXP( omega(1,4)-0.8d0)/(1+DEXP( omega(1,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,1)=omega(1,4)

omega(2,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,3))
    IF (omega(2,4)<-0.8) omega(2,4)=-0.8d0-((DEXP(-omega(2,4)-0.8d0)/(1+DEXP(-omega(2,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(2,4)> 0.8) omega(2,4)= 0.8d0+((DEXP( omega(2,4)-0.8d0)/(1+DEXP( omega(2,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,2)=omega(2,4)

omega(3,4)=DOT_PRODUCT(param(nthetab+nthetaa+nthetag+1:ntheta),xrtemp(:,1,2))
    IF (omega(3,4)<-0.8) omega(3,4)=-0.8d0-((DEXP(-omega(3,4)-0.8d0)/(1+DEXP(-omega(3,4)-0.8d0)))-0.5d0)*0.2d0
    IF (omega(3,4)> 0.8) omega(3,4)= 0.8d0+((DEXP( omega(3,4)-0.8d0)/(1+DEXP( omega(3,4)-0.8d0)))-0.5d0)*0.2d0
omega(4,3)=omega(3,4)

CASE DEFAULT
!    PRINT *,'nchild(m): ', nchild(m)
!    PRINT *,'error'
!    PAUSE
END SELECT

lchol=CHOL(omega)

END FUNCTION lcholbase





FUNCTION var_cov(param,n)
USE Globvar
! This function computes the VCov matrix of the estimates
! by using the numerical gradient of a function using forward differences
IMPLICIT NONE
INTEGER, INTENT(IN) :: n
REAL(8), INTENT(IN) :: param(n)
REAL(8) :: var_cov(n,n)

REAL(8) :: a(n,n)
REAL(8) :: f0			! Initial value of lnmlobj.
REAL(8) :: grdd(n),dh(n), arg(n,n)
REAL(8) :: ax0(n),xdh(n)
REAL(8) :: dax0(n),one
REAL(8) :: indgrad(n)
REAL(8) :: atemp(nobs)
INTEGER :: i, j, k

! Set up the gradient calculation
one  = 1.0D0	! This parameter affects the length of dh
grdd = 0.0D0
	! Computation of stepsize (dh) for gradient
ax0  = ABS(param)
dax0 = 1.0D0
WHERE (param.ne.0.0D0) dax0 = param/ax0
dh = (1.0D-6)*(1.0D-2)*one*dax0
WHERE (ax0.gt.(1.0D-2)*one) dh = (1.0D-8)*ax0*dax0
xdh = param+dh

! Calculate VCOV
a=0.0d0

atemp=astar

!!$OMP PARALLEL DO SCHEDULE(STATIC) PRIVATE(f0,grdd,arg,indgrad) 
	! For this part, parallezation does not work because f0, j, arg, etc have inter-dependence.
DO i = 1, nfam

arg = SPREAD(param,DIM=2,NCOPIES=n)

SELECT CASE(nchild(i))
CASE(1)
	CALL LIKELIHOOD1(f0,atemp(famindex(i)),i,param,n, &
		nthetab, nthetaa, xb(:,famindex(i)), xa(:,famindex(i)), sa, near(i))
CASE(2)
	CALL LIKELIHOOD2(f0,atemp(famindex(i)-1:famindex(i)),i,param,n, &
		ns,draws2seed,lcholbase(i,param,n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-1:famindex(i)), xa(:,famindex(i)-1:famindex(i)), xg(:,famindex(i)-1:famindex(i)), sa, sg, near(i))
CASE(3)
	CALL LIKELIHOOD3(f0,atemp(famindex(i)-2:famindex(i)),i,param,n, &
		ns,draws3seed,lcholbase(i,param,n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-2:famindex(i)), xa(:,famindex(i)-2:famindex(i)), xg(:,famindex(i)-2:famindex(i)), sa, sg, near(i))
CASE(4)
	CALL LIKELIHOOD4(f0,atemp(famindex(i)-3:famindex(i)),i,param,n, &
		ns,draws4seed,lcholbase(i,param,n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-3:famindex(i)), xa(:,famindex(i)-3:famindex(i)), xg(:,famindex(i)-3:famindex(i)), sa, sg, near(i))
CASE DEFAULT
    PRINT *,'error'
    PAUSE
END SELECT

		DO j = 1, n
			arg(j,j)=xdh(j)

        SELECT CASE(nchild(i))
        CASE(1)
			CALL LIKELIHOOD1(grdd(j),atemp(famindex(i)),i,arg(:,j),n, &
				nthetab, nthetaa, xb(:,famindex(i)), xa(:,famindex(i)), sa, near(i))
        CASE(2)
			CALL LIKELIHOOD2(grdd(j),atemp(famindex(i)-1:famindex(i)),i,arg(:,j),n, &
				ns,draws2seed,lcholbase(i,arg(:,j),n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-1:famindex(i)), xa(:,famindex(i)-1:famindex(i)), xg(:,famindex(i)-1:famindex(i)), sa, sg, near(i))
        CASE(3)
			CALL LIKELIHOOD3(grdd(j),atemp(famindex(i)-2:famindex(i)),i,arg(:,j),n, &
				ns,draws3seed,lcholbase(i,arg(:,j),n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-2:famindex(i)), xa(:,famindex(i)-2:famindex(i)), xg(:,famindex(i)-2:famindex(i)), sa, sg, near(i))
        CASE(4)
			CALL LIKELIHOOD4(grdd(j),atemp(famindex(i)-3:famindex(i)),i,arg(:,j),n, &
				ns,draws4seed,lcholbase(i,arg(:,j),n,nchild(i)),nthetab, nthetaa, nthetag, xb(:,famindex(i)-3:famindex(i)), xa(:,famindex(i)-3:famindex(i)), xg(:,famindex(i)-3:famindex(i)), sa, sg, near(i))
        CASE DEFAULT
            PRINT *,'error'
            PAUSE
        END SELECT


		END DO

	indgrad= (grdd-f0)/dh

	a=a+outerprod(indgrad,indgrad)

END DO
!!$OMP END PARALLEL DO

print *,'check'


DO j=1,ntheta
DO i=1,ntheta
	print *,'a(i,j):  ',i,j, a(i,j)
END DO
END DO

print *,'inverting var cov matrix'
var_cov=MATINV(a)

END FUNCTION var_cov




FUNCTION FAMSOLVE(param,n)
	! This function calculates simulated GHK probabilities for all the location
	! patterns by the number of children to evaluate the goodness-of-fit.
	! A standard way to do this would be to use a conventional frequency simulator.
	! But here it is based on GHK Monte Carlo integration, i.e. make simulation 
	! draws for each possible location pattern 
	! to compute its probability, and repeat this for every location pattern, based on
	! the equilibrium strategy profile, astar().

USE Globvar
USE PROBABILITY
USE GHK
IMPLICIT NONE
INTEGER, INTENT(IN) :: n 
REAL(8), INTENT(IN) :: param(n)

REAL(8) :: FAMSOLVE(0:15,4)
INTEGER :: i, j, k, l, m 
REAL(8) :: temp
REAL(8) :: ghkval		! The calculated GHK value
REAL(8) :: configprob(0:15)
INTEGER :: predconfig      ! predicted location configuration


FAMSOLVE=0.0d0

DO m=1,nfam

k=famindex(m)
ghkval=0.0d0
configprob=0.0d0

SELECT CASE(nchild(m))

CASE(1)

    configprob(1)=1.0d0-NORMCDF(-astar(k),0.0d0,1.0d0)
    configprob(0)=      NORMCDF(-astar(k),0.0d0,1.0d0)
    FAMSOLVE(1,1)=FAMSOLVE(1,1)+configprob(1)
    FAMSOLVE(0,1)=FAMSOLVE(0,1)+configprob(0)



CASE(2)

	DO i=0,3
		ghkval=GHK2(param,n,ns,draws2seed,lcholbase(m,param,n,nchild(m)),astar(famindex(m)-1:famindex(m)),i)
		FAMSOLVE(i,2)=FAMSOLVE(i,2)+ghkval
		configprob(i)=ghkval
	ENDDO


CASE(3)

	DO i=0,7
		ghkval=GHK3(param,n,ns,draws3seed,lcholbase(m,param,n,nchild(m)),astar(famindex(m)-2:famindex(m)),i)
		FAMSOLVE(i,3)=FAMSOLVE(i,3)+ghkval
		configprob(i)=ghkval
	ENDDO


CASE(4)

	DO i=0,15
		ghkval=GHK4(param,n,ns,draws4seed,lcholbase(m,param,n,nchild(m)),astar(famindex(m)-3:famindex(m)),i)
		FAMSOLVE(i,4)=FAMSOLVE(i,4)+ghkval
		configprob(i)=ghkval
	ENDDO

CASE DEFAULT
    PRINT *,'nchild(m): ', nchild(m)

    PRINT *,'error'
    PAUSE
END SELECT


temp=0.0d0
DO i=0,15
    IF (configprob(i)>temp) THEN
        temp=configprob(i)
        predconfig=i
    ENDIF
ENDDO
DO l=1,nchild(m)
    predobs(k-nchild(m)+l)=IBITS(predconfig,nchild(m)-l,1)
ENDDO


END DO  ! nfam



END FUNCTION FAMSOLVE









END MODULE Object









PROGRAM MAIN
USE Globvar
USE INIT
USE MINIMIZATION
USE Object
IMPLICIT NONE

REAL(8) :: resultmlv, resultmlse
REAL(8) :: gtol
REAL(4) :: time0, iniadj
INTEGER :: i, j, k, l, m, n, io
CHARACTER(12) :: horz(3)
CHARACTER(20) :: flname
CHARACTER(80) :: flname2
CHARACTER(20) :: fmtname           ! To accomodate different number of parameters
CHARACTER(50) :: inivalnote
CHARACTER(60) :: setupnote

REAL(8), ALLOCATABLE :: initheta(:)
REAL(8), ALLOCATABLE :: vcov(:,:)
REAL(8), ALLOCATABLE :: se(:)
REAL(8), ALLOCATABLE :: tval(:)
CHARACTER(16), ALLOCATABLE :: vert(:)
REAL(8), ALLOCATABLE :: pred(:,:)

horz(1)='        coef' ; horz(2)='        s.e.' ; horz(3)=' asy t-val'

! Read the setting file
    flname2='setting.prn'
    OPEN(1,file=(flname2))
    READ(1,fmt='(A60)') setupnote
    READ(1,fmt=*) nthetab
    READ(1,fmt=*) nthetaa
    READ(1,fmt=*) nthetag
    READ(1,fmt=*) nthetar
    READ(1,fmt=*) ntheta
    READ(1,fmt=*) nobs
    READ(1,fmt=*) nfam
    READ(1,fmt=*) ns
ALLOCATE (initheta(ntheta))
    READ(1,fmt=*) initheta
ALLOCATE (vert(ntheta))
    vert(1)='cons      '
    READ(1,fmt=*) vert(2:ntheta)
    CLOSE(1)




ALLOCATE (theta(ntheta))
ALLOCATE (vcov(ntheta,ntheta))
ALLOCATE (se(ntheta))
ALLOCATE (tval(ntheta))
ALLOCATE (famid(nobs))
ALLOCATE (famindex(nfam))
ALLOCATE (nchild(nfam))
ALLOCATE (y(nobs))
ALLOCATE (astar(nobs))
ALLOCATE (xb(nthetab,nobs))
IF (nthetaa>=3) THEN
	sa=nthetaa-2
	ALLOCATE (xa(nthetaa-2,nobs))
ELSEIF (nthetaa==1) THEN
	sa=1
	ALLOCATE (xa(1,nobs))
ENDIF
IF (nthetag>=2) THEN
	sg=nthetag-1
	ALLOCATE (xg(nthetag-1,nobs))
ELSEIF (nthetag==1) THEN
	sg=1
	ALLOCATE (xg(1,nobs))
ENDIF
ALLOCATE (xr(nthetar,nchildmax,nobs))
ALLOCATE (near(nfam))
ALLOCATE (predobs(nobs))
ALLOCATE (pred(0:2**nchildmax-1,4))


! Set gradient tolerance
gtol=0.000001d0

time0 = SECNDS(0.0)

CALL GETTIM(i, j, k, l)
CALL GETDAT(l, m, n)
flname2=CHAR(ICHAR('0')+(m/10))//CHAR(ICHAR('0')+(m-m/10*10))//CHAR(ICHAR('0')+(n/10))//CHAR(ICHAR('0')+(n-n/10*10))//CHAR(ICHAR('0')+(i/10))//CHAR(ICHAR('0')+(i-i/10*10))//CHAR(ICHAR('0')+(j/10))//CHAR(ICHAR('0')+(j-j/10*10))//CHAR(ICHAR('0')+(k/10))//CHAR(ICHAR('0')+(k-k/10*10))//'.prn'

OPEN(4,file=(flname2),STATUS='REPLACE')
WRITE(4,fmt='(A20,4x,A60)') 'File Name: ', flname2
WRITE(4,fmt='(A20,4x,A60)') 'Estimation Setting: ', setupnote
WRITE(4,fmt=*) '# of families     :  ', nfam
WRITE(4,fmt=*) '# of observations:  ', nobs
WRITE(4,fmt=*) '# of simuln draws  :  ', ns
WRITE(4,fmt=*) ' '

theta=initheta
numeval=0
astar=0.0d0

! Load the data set
CALL READDATA(nobs)


ALLOCATE (draws2seed(ns,2))
ALLOCATE (draws3seed(ns,3))
ALLOCATE (draws4seed(ns,4))
CALL RANDOM_SEED(PUT=(/3000/))
CALL RANDOM_NUMBER (draws2seed)
CALL RANDOM_NUMBER (draws3seed)
CALL RANDOM_NUMBER (draws4seed)
draws2seed(ns/2+1:ns,:)=1-draws2seed(1:ns/2,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:)=1-draws3seed(1:ns/2,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:)=1-draws4seed(1:ns/2,:)        ! This line makes antithetics

ALLOCATE (draws2e(ns,2))
ALLOCATE (draws3e(ns,3))
ALLOCATE (draws4e(ns,4))
CALL RANDOM_NUMBER (draws2e)
CALL RANDOM_NUMBER (draws3e)
CALL RANDOM_NUMBER (draws4e)
draws2seed(ns/2+1:ns,:)=1-draws2e(1:ns/2,:)        ! This line makes antithetics
draws3seed(ns/2+1:ns,:)=1-draws3e(1:ns/2,:)        ! This line makes antithetics
draws4seed(ns/2+1:ns,:)=1-draws4e(1:ns/2,:)        ! This line makes antithetics


! the following lines compute the estimates using a Quasi-Newton method
! with a numerical gradient

objsaved=0.0d0
CALL dfpmin(theta,ntheta,gtol,lnmlobj)
objsaved=lnmlobj(theta,ntheta,0)
objsaved=-objsaved


print *,'1st round Minimization Finished!'
print *,'objsaved : ',objsaved

print *,'Calculating variance-covariance matrix...'

vcov=var_cov(theta, ntheta)

print *,'Vcov obtained!'
!	print *,vcov


DO i=1,ntheta
	print *,'vcov(i,i):  ',i, vcov(i,i)
END DO

DO i=1,ntheta
	se(i)=DSQRT(vcov(i,i))
	tval(i)=theta(i)/se(i)
END DO




WRITE(4,fmt=*) '# of parameters:  ', ntheta
WRITE(4,fmt=*) ' '
WRITE(4,fmt='(10x,3A16)') (horz(i),i=1,3)
WRITE(4,fmt=*) ' '
DO i=1,ntheta
	WRITE(4,fmt='(A16,3F16.10)') vert(i), theta(i), se(i), tval(i)
ENDDO

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Data'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
DO i=0,2**nchildmax-1
    DO j=1,nfam
        IF (near(j)==i) THEN
            pred(i,nchild(j))=pred(i,nchild(j))+1
        ENDIF
    ENDDO
ENDDO
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)

pred=0
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Location patterns: Predicted'
WRITE(4,fmt='(A8,10A8)') '  config',' 1 child','     (%)',' 2 child','     (%)',' 3 child','     (%)',' 4 child','     (%)'
    pred(:,:)=FAMSOLVE(theta,ntheta)
DO i=0,2**nchildmax-1
	WRITE(4,fmt='(4X,4I1,4(F8.1,F8.2))') IBITS(i,3,1),IBITS(i,2,1),IBITS(i,1,1),IBITS(i,0,1), (pred(i,k),(pred(i,k)/SUM(pred(:,k))*100.0d0),k=1,4)
ENDDO
WRITE(4,fmt='(A8,4(F8.1,F8.2))') '   total', ((SUM(pred(:,j)),100.0),j=1,4)



WRITE(4,fmt=*) ' '
n=0
DO m=1,nfam
DO j=1,nchild(m)
    IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) n=n+1
ENDDO
ENDDO
WRITE(4,fmt=*) '% of observations correctly predicted:  ', n*1.0d0/nobs*100
n=0
DO m=1,nfam
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
ENDDO
WRITE(4,fmt=*) '% of families correctly predicted:  ', n*1.0d0/nfam*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==1) THEN
        IF (predobs(famindex(m))==IBITS(near(m),0,1)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 1 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==2) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 2 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==3) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 3 child families correctly predicted:  ', n*1.0d0/l*100
n=0; l=0
DO m=1,nfam
    IF (nchild(m)==4) THEN
    k=0
    DO j=1,nchild(m)
        IF (predobs(famindex(m)-nchild(m)+j)==IBITS(near(m),nchild(m)-j,1)) k=k+1
    ENDDO
    IF (k==nchild(m)) n=n+1
    l=l+1
    ENDIF
ENDDO
WRITE(4,fmt=*) '% of 4 child families correctly predicted:  ', n*1.0d0/l*100

WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'ML objective:  ', objsaved
WRITE(4,fmt=*) '# of objective function evaluations:  ', numeval
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'run time(seconds):  ', SECNDS(time0)
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) ' '
WRITE(4,fmt=*) 'Initial Values: '
DO i=1,ntheta
WRITE(4,fmt='(F16.10)') initheta(i)
END DO
WRITE(4,fmt=*) ' '

CLOSE(4)
CLOSE(5)

! estimated parameters
OPEN(1,file='param1.out',STATUS='REPLACE')
OPEN(2,file='param2.out',STATUS='REPLACE')
flname='('//CHAR((nthetab)/100+48)//CHAR(MOD(nthetab,100)/10+48)//CHAR(MOD(nthetab,10)+48)//'F20.14)'
WRITE(1,fmt=flname) theta(1:nthetab)
IF (ntheta>nthetab) THEN
flname='('//CHAR((ntheta-nthetab)/100+48)//CHAR(MOD(ntheta-nthetab,100)/10+48)//CHAR(MOD(ntheta-nthetab,10)+48)//'F20.14)'
WRITE(2,fmt=flname) theta(nthetab+1:ntheta)
ENDIF
CLOSE(1)
CLOSE(2)

END PROGRAM MAIN
